from himena.widgets._main_window import MainWindow, BackendMainWindow
from himena.widgets._widget_list import TabArea
from himena.widgets._wrapper import SubWindow, ParametricWindow, DockWidget
from himena.widgets._initialize import (
    current_instance,
    set_current_instance,
    remove_instance,
    init_application,
)
from himena.widgets._functions import (
    set_status_tip,
    get_clipboard,
    set_clipboard,
    notify,
    append_result,
)

__all__ = [
    # widgets
    "MainWindow",
    "TabArea",
    "BackendMainWindow",
    "SubWindow",
    "ParametricWindow",
    "DockWidget",
    # functions
    "current_instance",
    "set_current_instance",
    "remove_instance",
    "init_application",
    "set_status_tip",
    "get_clipboard",
    "set_clipboard",
    "notify",
    "append_result",
]



import time
from typing import Annotated
import numpy as np
from himena.consts import StandardType
from himena.plugins import register_function, configure_gui
from himena.types import Parametric, WidgetDataModel, WidgetConstructor
from himena.widgets import MainWindow, notify, set_status_tip
from himena_builtins.debug import draw_tool

del draw_tool

TOOLS_DEBUG = "tools/debug"
TOOLS_DEBUG_ERROR_WARN = "tools/debug/error_warn"
TOOLS_DEBUG_ASYNC = "tools/debug/async"


@register_function(
    menus=TOOLS_DEBUG_ERROR_WARN,
    title="Just raise an exception",
    command_id="debug:raise-exception",
)
def raise_exception() -> Parametric:
    def run(a: int = 0):
        raise ValueError("This is a test exception")

    return run


@register_function(
    menus=TOOLS_DEBUG_ERROR_WARN,
    title="Just raise an exception (async)",
    run_async=True,
    command_id="debug:raise-exception-async",
)
def raise_exception_async() -> Parametric:
    def run(a: int = 0):
        raise ValueError("This is a test exception")

    return run


@register_function(
    menus=TOOLS_DEBUG_ERROR_WARN,
    title="Just warn",
    command_id="debug:warning",
)
def raise_warning():
    import warnings

    warnings.warn("This is a test warning", UserWarning, stacklevel=2)


@register_function(
    menus=TOOLS_DEBUG_ERROR_WARN,
    title="Just warn (async)",
    run_async=True,
    command_id="debug:warning-async",
)
def raise_warning_async() -> Parametric:
    import warnings

    def run():
        warnings.warn("This is a test warning", UserWarning, stacklevel=2)

    return run


@register_function(
    menus=TOOLS_DEBUG,
    title="Raise when warning",
    command_id="debug:raise-when-warning",
)
def raise_when_warning():
    import warnings

    warnings.simplefilter("error")


@register_function(
    menus=TOOLS_DEBUG,
    title="Test model drop",
    command_id="debug:test-model-drop",
)
def function_with_model_drop() -> Parametric:
    def run(
        model_any: WidgetDataModel,
        model_text: Annotated[WidgetDataModel[str], {"types": "text"}],
        model_image: Annotated[WidgetDataModel, {"types": "array.image"}],
    ):
        print(model_any)
        print(model_text)
        print(model_image)

    return run


@register_function(
    menus=TOOLS_DEBUG,
    title="Open User Preferences",
    command_id="debug:open-user-preferences",
)
def open_user_preferences(ui: MainWindow):
    from himena.profile import data_dir, profile_dir

    output = []
    for path in profile_dir().glob("*.json"):
        output.append(path)
    output.append(data_dir() / "recent.json")
    output.append(data_dir() / "recent_sessions.json")
    ui.read_files(output)
    return None


@register_function(
    menus=TOOLS_DEBUG,
    title="Test preview",
    command_id="debug:test-preview",
)
def preview_test() -> Parametric:
    @configure_gui(preview=True)
    def testing_preview(a: int, b: str, is_previewing: bool = False) -> WidgetDataModel:
        out = f"a = {a!r}\nb ={b!r}"
        if is_previewing:
            out += "\n(preview)"
        print(f"called with {a=}, {b=}, {is_previewing=}")
        return WidgetDataModel(value=out, type=StandardType.TEXT)

    return testing_preview


@register_function(
    menus=TOOLS_DEBUG,
    title="Test plot",
    command_id="debug:test-plot",
)
def plot_test() -> Parametric:
    import himena.standards.plotting as hplt

    @configure_gui(preview=True)
    def run(a: int, b: int = 4) -> WidgetDataModel:
        fig = hplt.figure()
        fig.axes.plot([0, 1, 2], [2, a, b], color="red")
        fig.axes.title = "Test plot"
        return WidgetDataModel(value=fig, type=StandardType.PLOT)

    return run


@register_function(
    menus=TOOLS_DEBUG,
    title="Test plot with parametric (below)",
    command_id="debug:test-plot-parametric-below",
)
def plot_test_parametric_below() -> Parametric:
    import himena.standards.plotting as hplt

    @configure_gui(preview=True, result_as="below")
    def run(freq: float = 1.0, phase: float = 0.0) -> WidgetDataModel:
        fig = hplt.figure()
        x = np.linspace(0, 2 * np.pi, 100)
        y = np.sin(freq * x + phase)
        fig.axes.plot(x, y, color="red")
        fig.axes.title = "Test plot"
        return WidgetDataModel(value=fig, type=StandardType.PLOT)

    return run


@register_function(
    menus=TOOLS_DEBUG,
    title="Test plot with parametric (right)",
    command_id="debug:test-plot-parametric-right",
)
def plot_test_parametric_right() -> Parametric:
    import himena.standards.plotting as hplt

    @configure_gui(preview=True, result_as="right")
    def run(freq: float = 1.0, phase: float = 0.0) -> WidgetDataModel:
        fig = hplt.figure()
        x = np.linspace(0, 2 * np.pi, 100)
        y = np.sin(freq * x + phase)
        fig.axes.plot(x, y, color="red")
        fig.axes.title = "Test plot"
        return WidgetDataModel(value=fig, type=StandardType.PLOT)

    return run


@register_function(
    menus=TOOLS_DEBUG_ASYNC,
    title="Test async checkpoints (no param)",
    run_async=True,
    command_id="debug:test-async-checkpoints-no-param",
)
def test_async_checkpoints_no_param() -> WidgetDataModel:
    import time

    for i in range(10):
        time.sleep(0.3)
    return WidgetDataModel(value="Done", type=StandardType.TEXT)


@register_function(
    menus=TOOLS_DEBUG_ASYNC,
    title="Test async checkpoints",
    command_id="debug:test-async-checkpoints",
    run_async=True,
)
def test_async_checkpoints() -> Parametric:
    import time

    def run():
        for i in range(10):
            time.sleep(0.3)
        return

    return run


@register_function(
    menus=TOOLS_DEBUG_ASYNC,
    title="Test async widget creation",
    run_async=True,
    command_id="debug:test-async-widget-creation",
)
def test_async_widget_creation() -> Parametric:
    from magicgui.widgets import Container, LineEdit

    def make_widget(texts: list[str]):
        con = Container()
        for text in texts:
            con.append(LineEdit(value=text))
        return con.native

    def run() -> WidgetConstructor:
        texts = []
        for i in range(10):
            time.sleep(0.3)
            texts.append(f"Text {i}")
        return lambda: make_widget(texts)

    return run


@register_function(
    menus=TOOLS_DEBUG,
    title="Test notification",
    run_async=True,
    command_id="debug:test-notification",
)
def test_notification() -> Parametric:
    def run(notification: bool = True, status_tip: bool = True):
        time.sleep(1)
        if status_tip:
            notify("1. This is test notification", duration=4.2)
        if notification:
            set_status_tip("1. This is test status tip", duration=4.2)
        time.sleep(1)
        if status_tip:
            notify("2. This is test notification", duration=4.2)
        if notification:
            set_status_tip("2. This is test status tip", duration=4.2)

    return run



from __future__ import annotations
from typing import Generic, TypeVar

from magicgui.widgets import LineEdit
from magicgui.widgets.bases import ValueWidget
from magicgui.types import Undefined
from magicgui.application import use_app
from magicgui.backends._qtpy.widgets import QBaseStringWidget
from himena.qt._qlineedit import (
    QIntLineEdit,
    QDoubleLineEdit,
    QCommaSeparatedIntLineEdit,
    QCommaSeparatedDoubleLineEdit,
)

__all__ = ["IntEdit", "FloatEdit"]


class QBaseRangedStringWidget(QBaseStringWidget):
    def _mgui_get_min(self) -> float:
        """Get the minimum possible value."""
        val = self._qwidget.minimum()
        return val

    def _mgui_set_min(self, value: float):
        """Set the minimum possible value."""
        self._qwidget.setMinimum(value)

    def _mgui_get_max(self) -> float:
        """Set the maximum possible value."""
        return self._qwidget.maximum()

    def _mgui_set_max(self, value: float):
        """Set the maximum possible value."""
        self._qwidget.setMaximum(value)


_T = TypeVar("_T")


class RangedLineEdit(LineEdit, Generic[_T]):
    def __init__(
        self,
        value=Undefined,
        min=Undefined,
        max=Undefined,
        widget_type=None,
        **kwargs,
    ):
        app = use_app()
        assert app.native
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=widget_type,
            **kwargs,
        )
        if min is not Undefined:
            self.min = min
        if max is not Undefined:
            self.max = max
        self._set_nullable(self._nullable)

    @property
    def min(self) -> _T:
        return self._widget._mgui_get_min()

    @min.setter
    def min(self, value: _T):
        self._widget._mgui_set_min(value)

    @property
    def max(self) -> _T:
        return self._widget._mgui_get_max()

    @max.setter
    def max(self, value: _T):
        self._widget._mgui_set_max(value)

    def _set_nullable(self, nullable: bool):
        self._widget._qwidget._empty_allowed = nullable


class QIntEdit(QBaseRangedStringWidget):
    _qwidget: QIntLineEdit

    def __init__(self, **kwargs) -> None:
        super().__init__(QIntLineEdit, "text", "setText", "valueChanged", **kwargs)

    def _post_get_hook(self, value):
        if value == "":
            return None
        return int(value)

    def _pre_set_hook(self, value):
        return str(value)


class IntEdit(RangedLineEdit):
    def __init__(
        self,
        value=Undefined,
        min=Undefined,
        max=Undefined,
        **kwargs,
    ):
        super().__init__(
            value=value,
            min=min,
            max=max,
            widget_type=QIntEdit,
            **kwargs,
        )

    def get_value(self) -> int:
        val = super().get_value()
        if val is None and not self._nullable:
            raise ValueError(f"Must specify a value for {self.label!r}")
        return val

    def set_value(self, value):
        if value is None:
            if not self._nullable:
                raise ValueError(f"Value for {self.label} cannot be None")
            value = ""
        super().set_value(value)


class QFloatEdit(QBaseRangedStringWidget):
    _qwidget: QDoubleLineEdit

    def __init__(self, **kwargs) -> None:
        super().__init__(QDoubleLineEdit, "text", "setText", "valueChanged", **kwargs)

    def _post_get_hook(self, value):
        if value == "":
            return None
        return float(value)

    def _pre_set_hook(self, value):
        return str(value)


class FloatEdit(RangedLineEdit):
    def __init__(
        self,
        value=Undefined,
        min=Undefined,
        max=Undefined,
        **kwargs,
    ):
        super().__init__(
            value=value,
            min=min,
            max=max,
            widget_type=QFloatEdit,
            **kwargs,
        )

    def get_value(self) -> float:
        val = super().get_value()
        if val is None and not self._nullable:
            raise ValueError(f"Must specify a value for {self.label}")
        return val

    def set_value(self, value):
        if value is None:
            if not self._nullable:
                raise ValueError(f"Value for {self.label} cannot be None")
            value_str = ""
        else:
            value_str = float_to_str(value)
        super().set_value(value_str)


class QIntListEdit(QBaseStringWidget):
    _qwidget: QCommaSeparatedIntLineEdit

    def __init__(self, **kwargs) -> None:
        super().__init__(
            QCommaSeparatedIntLineEdit, "text", "setText", "textChanged", **kwargs
        )


class IntListEdit(LineEdit):
    def __init__(self, value=Undefined, **kwargs):
        app = use_app()
        assert app.native
        kwargs["widget_type"] = QIntListEdit
        ValueWidget.__init__(
            self,
            value=value,
            **kwargs,
        )

    def get_value(self) -> list[int]:
        val = super().get_value()
        if val is None and not self._nullable:
            raise ValueError(f"Must specify a value for {self.label}")
        if val.strip() == "":
            return []
        return [int(part) for part in val.split(",")]

    def set_value(self, value):
        if value is None:
            value_str = ""
        else:
            value_str = ", ".join(str(part) for part in value)
        super().set_value(value_str)


class QFloatListEdit(QBaseStringWidget):
    _qwidget: QCommaSeparatedDoubleLineEdit

    def __init__(self, **kwargs) -> None:
        super().__init__(
            QCommaSeparatedDoubleLineEdit, "text", "setText", "textChanged", **kwargs
        )


class FloatListEdit(LineEdit):
    def __init__(self, value=Undefined, **kwargs):
        app = use_app()
        assert app.native
        kwargs["widget_type"] = QFloatListEdit
        ValueWidget.__init__(
            self,
            value=value,
            **kwargs,
        )

    def get_value(self) -> list[float]:
        val = super().get_value()
        if val is None and not self._nullable:
            raise ValueError(f"Must specify a value for {self.label}")
        if val.strip() == "":
            return []
        return [float(part) for part in val.split(",")]

    def set_value(self, value):
        if value is None:
            value_str = ""
        else:
            value_str = ",".join(float_to_str(part) for part in value)
        super().set_value(value_str)


def float_to_str(value: int | float):
    if isinstance(value, int) or hasattr(value, "__index__"):
        value_str = str(value)
        if len(value_str) > 5:
            value_str = format(value, ".8g")
        return value_str
    out = format(value, ".8g")
    if "." not in out:
        return f"{out}.0"
    return out



from __future__ import annotations

from typing import (
    Annotated,
    Callable,
    Literal,
    TypeVar,
    get_origin,
    get_args,
    Any,
    overload,
)
import inspect
from himena.types import GuiConfiguration


def _is_annotated(annotation: Any) -> bool:
    """Check if a type hint is an Annotated type."""
    return get_origin(annotation) is Annotated


def _split_annotated_type(annotation) -> tuple[Any, dict]:
    """Split an Annotated type into its base type and options dict."""
    if not _is_annotated(annotation):
        raise TypeError("Type hint must be an 'Annotated' type.")

    typ, *meta = get_args(annotation)
    all_meta = {}
    for m in meta:
        if not isinstance(m, dict):
            raise TypeError(
                "Invalid Annotated format for magicgui. Arguments must be a dict"
            )
        all_meta.update(m)

    return typ, all_meta


_F = TypeVar("_F", bound=Callable)


@overload
def configure_gui(
    f: _F,
    *,
    title: str | None = None,
    preview: bool = False,
    auto_close: bool = True,
    show_parameter_labels: bool = True,
    gui_options: dict[str, Any] | None = None,
    result_as: Literal["window", "below", "right"] = "window",
    **kwargs,
) -> _F: ...
@overload
def configure_gui(
    *,
    title: str | None = None,
    preview: bool = False,
    auto_close: bool = True,
    show_parameter_labels: bool = True,
    gui_options: dict[str, Any] | None = None,
    result_as: Literal["window", "below", "right"] = "window",
    **kwargs,
) -> Callable[[_F], _F]: ...


def configure_gui(
    f=None,
    *,
    title: str | None = None,
    preview: bool = False,
    auto_close: bool = True,
    show_parameter_labels: bool = True,
    gui_options: dict[str, Any] | None = None,
    result_as: Literal["window", "below", "right"] = "window",
    **kwargs,
):
    """Configure the parametric GUI.

    This decorator sets the configuration options for the parametric GUI window.

    ``` python
    @configure_gui(a={"label": "A", "widget_type": "FloatSlider"})
    def my_func(a: float):
        pass
    ```

    Parameters
    ----------
    title : str, optional
        The title of the parametric GUI window. If not provided, this title will be
        determined by the action title where this function is returned.
    preview : bool, default False
        If true, a preview toggle switch will be added to the GUI window. When the
        switch is on, the function will be called and the result will be displayed. Note
        that `configure_gui` does not consider whether the preview is a heavy operation.
    auto_close : bool, default True
        If true, the parametric GUI window will be closed automatically after the
        function is executed.
    show_parameter_labels : bool, default True
        If true, the parameter names will be shown in the GUI window.
    gui_options : dict, optional
        Additional GUI options to be passed to the `magicgui` decorator. Keys can also
        be passed as variable keyword arguments **kwargs.
    """
    kwargs = dict(**kwargs, **(gui_options or {}))

    def _inner(f):
        sig = inspect.signature(f)
        new_params = sig.parameters.copy()
        if var_kwargs_name := _get_var_kwargs_name(sig):
            new_params.pop(var_kwargs_name)

        for k, v in kwargs.items():
            if k not in new_params:
                if var_kwargs_name is None:
                    raise TypeError(f"{k!r} is not a valid parameter for {f!r}.")
                # This allows using **kwargs in the target function so that magicgui
                # widget can be created for a variable number of parameters.
                param = inspect.Parameter(name=k, kind=inspect.Parameter.KEYWORD_ONLY)
            else:
                param = sig.parameters[k]
            if isinstance(v, dict) and "annotation" in v:
                param_annotation = v.pop("annotation")
            else:
                param_annotation = param.annotation
            # unwrap Annotated types
            if not _is_annotated(param_annotation):
                annot = _prioritize_choices(param_annotation, v)
                param = param.replace(annotation=Annotated[annot, v])
            else:
                typ, meta = _split_annotated_type(param_annotation)
                meta.update(v)
                typ = _prioritize_choices(typ, meta)
                param = param.replace(annotation=Annotated[typ, meta])
            new_params[k] = param
        # update the signature with the normalize one
        sig = sig.replace(parameters=list(new_params.values()))
        f.__signature__ = sig
        f.__annotations__ = {k: v.annotation for k, v in sig.parameters.items()}
        if sig.return_annotation is not inspect.Parameter.empty:
            f.__annotations__["return"] = sig.return_annotation

        GuiConfiguration(
            title=title,
            preview=preview,
            auto_close=auto_close,
            show_parameter_labels=show_parameter_labels,
            result_as=result_as,
        ).set(f)
        return f

    return _inner if f is None else _inner(f)


def _prioritize_choices(annotation, options: dict[str, Any]):
    if "choices" in options:
        typ = Any
    else:
        typ = annotation
    return typ


def _get_var_kwargs_name(sig: inspect.Signature) -> str | None:
    """Get the name of the **kwargs parameter."""
    for p in sig.parameters.values():
        if p.kind == inspect.Parameter.VAR_KEYWORD:
            return p.name
    return None



from __future__ import annotations

from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import contextmanager
from logging import getLogger
from pathlib import Path
import threading
import uuid
from typing import (
    Any,
    Callable,
    Sequence,
    Generic,
    Iterator,
    Literal,
    TypeVar,
    overload,
    TYPE_CHECKING,
)
import warnings
from app_model.expressions import create_context
from psygnal import SignalGroup, Signal

from himena._app_model import AppContext, HimenaApplication
from himena._open_recent import RecentFileManager, RecentSessionManager
from himena._utils import get_widget_class_id, import_object
from himena.consts import NO_RECORDING_FIELD
from himena.plugins import _checker, actions as _actions
from himena.profile import AppProfile, load_app_profile
from himena.style import Theme
from himena.standards import BaseMetadata
from himena.types import (
    ClipboardDataModel,
    Size,
    WidgetDataModel,
    NewWidgetBehavior,
    DockArea,
    DockAreaString,
    BackendInstructions,
    WindowRect,
)
from himena.utils.misc import is_subtype
from himena.widgets._backend import BackendMainWindow
from himena.widgets._hist import HistoryContainer, FileDialogHistoryDict
from himena.widgets._initialize import remove_instance
from himena.widgets._typemap import ObjectTypeMap, register_defaults
from himena.widgets._widget_list import TabList, TabArea, DockWidgetList
from himena.widgets._wrapper import ParametricWindow, SubWindow, DockWidget
from himena.workflow import ProgrammaticMethod
from himena._socket import SocketInfo

if TYPE_CHECKING:
    from app_model.types import KeyBindingRule
    from app_model.expressions import BoolOp
    from himena.widgets._widget_list import PathOrPaths

_W = TypeVar("_W")  # backend widget type
_T = TypeVar("_T")  # internal data type
_F = TypeVar("_F")  # function type
_LOGGER = getLogger(__name__)


class MainWindowEvents(SignalGroup, Generic[_W]):
    """Main window events."""

    tab_activated = Signal(TabArea[_W])
    window_activated = Signal(SubWindow[_W])


class MainWindow(Generic[_W]):
    """The main window handler object."""

    def __init__(
        self,
        backend: BackendMainWindow[_W],
        app: HimenaApplication,  # must be initialized
        theme: Theme,
    ) -> None:
        from himena.widgets._initialize import set_current_instance

        self._events = MainWindowEvents()
        self._backend_main_window = backend
        self._internal_clipboard_data: Any | None = None
        self._tab_list = TabList(backend)
        self._new_widget_behavior = NewWidgetBehavior.WINDOW
        self._model_app = app
        self._instructions = BackendInstructions()
        self._history_tab = HistoryContainer[int](max_size=20)
        self._history_command = HistoryContainer[str](max_size=200)
        self._history_closed = HistoryContainer[tuple[Path, str | None]](max_size=10)
        self._file_dialog_hist = FileDialogHistoryDict()
        app.commands.executed.connect(self._on_command_execution)
        backend._connect_main_window_signals(self)
        self._ctx_keys = AppContext(create_context(self, max_depth=0))
        self._tab_list.changed.connect(backend._update_context)
        self._dock_widget_list = DockWidgetList(backend)
        self._recent_manager = RecentFileManager.default(app)
        self._recent_session_manager = RecentSessionManager.default(app)
        if "." not in app.name:
            # likely a mock instance
            set_current_instance(app.name, self)
            self._recent_manager.update_menu()
            self._recent_session_manager.update_menu()
        self._executor = ThreadPoolExecutor(max_workers=5)
        self._global_lock = threading.Lock()
        self._object_type_map = ObjectTypeMap()
        self.theme = theme
        register_defaults(self._object_type_map)

    @property
    def events(self) -> MainWindowEvents[_W]:
        """Main window events."""
        return self._events

    @property
    def object_type_map(self) -> ObjectTypeMap:
        """Mapping object to string that describes the type."""
        return self._object_type_map

    @property
    def socket_info(self) -> SocketInfo:
        """Socket information."""
        eh = self._backend_main_window._event_loop_handler
        return SocketInfo(host=eh._host, port=eh._port)

    @property
    def theme(self) -> Theme:
        """Get the current color theme of the main window."""
        return self._theme

    @theme.setter
    def theme(self, theme: str | Theme) -> None:
        """Set the style of the main window."""
        if isinstance(theme, str):
            theme = Theme.from_global(theme)
        self._theme = theme
        self._backend_main_window._update_widget_theme(theme)

        # if child implements "theme_changed_callback", call it
        for win in self.iter_windows():
            _checker.call_theme_changed_callback(win.widget, theme)
        for dock in self.dock_widgets:
            _checker.call_theme_changed_callback(dock.widget, theme)
        return None

    @property
    def app_profile(self) -> AppProfile:
        """Get the current application profile object."""
        return load_app_profile(self._model_app.name)

    def submit_async_task(
        self,
        func: Callable,
        *args,
        progress_description: str | None = None,
        **kwargs,
    ) -> Future:
        """Submit a task to the thread pool.

        Parameters
        ----------
        func : callable
            Function to run in the background.
        progress_description : str, optional
            Description of the task in the progress bar.
        """
        future = self._executor.submit(func, *args, **kwargs)
        if progress_description is None:
            progress_description = f"Running {func!r}"
        self._backend_main_window._add_job_progress(
            future, desc=progress_description, total=0
        )
        self.model_app.injection_store.process(future)
        return future

    @property
    def tabs(self) -> TabList[_W]:
        """Tab list object."""
        return self._tab_list

    def windows_for_type(self, types: str | list[str]) -> list[SubWindow[_W]]:
        """Get all sub-windows for the given types."""
        windows = []
        if isinstance(types, str):
            types = [types]
        if tab := self.tabs.current():
            for win in tab:
                mtype = win.model_type()
                if mtype and any(is_subtype(mtype, t) for t in types):
                    windows.append(win)
        return windows

    @property
    def dock_widgets(self) -> DockWidgetList[_W]:
        """Dock widget list object."""
        return self._dock_widget_list

    @property
    def model_app(self) -> HimenaApplication:
        """The app-model application instance."""
        return self._model_app

    @property
    def area_size(self) -> tuple[int, int]:
        """(width, height) of the main window tab area."""
        return self._backend_main_window._area_size()

    @property
    def rect(self) -> WindowRect:
        """Window rect (left, top, width, height) of the main window."""
        return self._backend_main_window._main_window_rect()

    @rect.setter
    def rect(self, value) -> None:
        rect = WindowRect.from_tuple(*value)
        return self._backend_main_window._set_main_window_rect(rect)

    @property
    def size(self) -> Size:
        """Size (width, height) of the main window."""
        return self.rect.size()

    @size.setter
    def size(self, size) -> None:
        r0 = self.rect
        s0 = Size(*size)
        self.rect = WindowRect(r0.left, r0.top, s0.width, s0.height)

    @property
    def clipboard(self) -> ClipboardDataModel | None:
        """Get the clipboard data as a ClipboardDataModel instance."""
        model = self._backend_main_window._clipboard_data()
        model.internal_data = self._internal_clipboard_data
        return model

    @clipboard.setter
    def clipboard(self, data: str | ClipboardDataModel) -> None:
        """Set the clipboard data."""
        if isinstance(data, str):
            data = ClipboardDataModel(text=data)
        elif not isinstance(data, ClipboardDataModel):
            raise ValueError("Clipboard data must be a ClipboardDataModel instance.")
        _LOGGER.info("Setting clipboard data: %r", data)
        self._backend_main_window._set_clipboard_data(data)
        self._internal_clipboard_data = data.internal_data

    def set_clipboard(
        self,
        *,
        text: str | None = None,
        html: str | None = None,
        image: Any | None = None,
        files: list[str | Path] | None = None,
        internal_data: Any | None = None,
    ) -> None:
        """Set clipboard data."""
        self.clipboard = ClipboardDataModel(
            text=text,
            html=html,
            image=image,
            files=files or [],
            internal_data=internal_data,
        )

    def add_tab(self, title: str | None = None) -> TabArea[_W]:
        """Add a new tab of given name."""
        return self.tabs.add(title)

    def window_for_id(self, identifier: uuid.UUID) -> SubWindow[_W] | None:
        """Retrieve a sub-window by its identifier."""
        if not isinstance(identifier, uuid.UUID):
            raise ValueError(f"Expected UUID, got {identifier!r}.")
        for win in self.iter_windows():
            if win._identifier == identifier:
                return win

    def _window_for_workflow_id(self, identifier: uuid.UUID) -> SubWindow[_W] | None:
        """Retrieve a sub-window by its workflow identifier."""
        for win in self.iter_windows():
            try:
                last_id = win._widget_workflow.last_id()
            except Exception:
                return None
            if last_id == identifier:
                return win

    def _current_or_new_tab(self) -> tuple[int, TabArea[_W]]:
        if self._new_widget_behavior is NewWidgetBehavior.WINDOW:
            if len(self.tabs) == 0:
                self.add_tab()
                idx = 0
            else:
                idx = self._backend_main_window._current_tab_index()
            tabarea = self.tabs[idx]
        else:
            tabarea = self.add_tab()
            idx = len(self.tabs) - 1
        return idx, tabarea

    def add_widget(
        self,
        widget: _W,
        *,
        title: str | None = None,
    ) -> SubWindow[_W]:
        """Add a widget to the sub window.

        Any widget that can be interpreted by the backend can be added. For example, for
        Qt application, you can add any QWidget instance:

        ```python
        ui.add_widget(QtW.QLabel("Hello world!"), title="my widget!")
        ```

        Parameters
        ----------
        widget : Any
            Widget to add.
        title : str, optional
            Title of the sub-window. If not given, its name will be automatically
            generated.

        Returns
        -------
        SubWindow
            The sub-window handler.
        """
        _, tabarea = self._current_or_new_tab()
        return tabarea.add_widget(widget, title=title)

    def add_dock_widget(
        self,
        widget: _W,
        *,
        title: str | None = None,
        area: DockAreaString | DockArea | None = DockArea.RIGHT,
        allowed_areas: list[DockAreaString | DockArea] | None = None,
        _identifier: uuid.UUID | None = None,
    ) -> DockWidget[_W]:
        """Add a custom widget as a dock widget of the main window.

        Parameters
        ----------
        widget : Widget type
            Widget instance that is allowed for the backend.
        title : str, optional
            Title of the dock widget.
        area : dock area, default DockArea.RIGHT
            String or DockArea enum that describes where the dock widget initially
            appears.
        allowed_areas : list of dock area, optional
            List of allowed dock areas for the widget.

        Returns
        -------
        DockWidget
            The dock widget handler.
        """
        dock = DockWidget(widget, self._backend_main_window, identifier=_identifier)
        dock_native = dock._split_interface_and_frontend()[1]
        self._backend_main_window.add_dock_widget(
            dock_native, title=title, area=area, allowed_areas=allowed_areas
        )
        self._dock_widget_list._add_dock_widget(dock)
        _checker.call_widget_added_callback(widget)
        _checker.call_theme_changed_callback(widget, self.theme)
        return dock

    def add_object(
        self,
        value: Any,
        *,
        type: str | None = None,
        title: str | None = None,
        force_open_with: str | None = None,
        metadata: Any | None = None,
    ) -> SubWindow[_W]:
        """Add any data as a widget data model.

        Parameters
        ----------
        value : Any
            Any object. Whether it can be represented as a widget is dependent on the
            plugins that are installed.
        type : str, optional
            Any str that describes the type of the object. This type must be registered
            with a proper widget class.
        title : str, optional
            Title of the sub-window.

        Returns
        -------
        SubWindow
            The sub-window handler.
        """
        if type is None:
            if isinstance(metadata, BaseMetadata):
                type = metadata.expected_type()
        if type is None:
            type = self._object_type_map.pick_type(value)
        wd = WidgetDataModel(
            value=value,
            type=type,
            title=title,
            force_open_with=force_open_with,
            metadata=metadata,
            workflow=ProgrammaticMethod(output_model_type=type).construct_workflow(),
        )
        return self.add_data_model(wd)

    def add_data_model(self, model_data: WidgetDataModel) -> SubWindow[_W]:
        """Add a widget data model as a widget."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.add_data_model(model_data)

    def add_function(
        self,
        func: Callable[..., _T],
        *,
        preview: bool = False,
        title: str | None = None,
        show_parameter_labels: bool = True,
        auto_close: bool = True,
        run_async: bool = False,
        result_as: Literal["window", "below", "right"] = "window",
    ) -> ParametricWindow[_W]:
        """Add a function as a parametric sub-window.

        The input function must return a `WidgetDataModel` instance, which can be
        interpreted by the application.

        Parameters
        ----------
        func : function (...) -> WidgetDataModel
            Function that generates a model from the input parameters.
        title : str, optional
            Title of the sub-window.

        Returns
        -------
        SubWindow
            The sub-window instance that represents the output model.
        """
        _, tabarea = self._current_or_new_tab()
        return tabarea.add_function(
            func, title=title, preview=preview, result_as=result_as,
            show_parameter_labels=show_parameter_labels, auto_close=auto_close,
            run_async=run_async,
        )  # fmt: skip

    def add_parametric_widget(
        self,
        widget: _W,
        callback: Callable | None = None,
        *,
        title: str | None = None,
        preview: bool = False,
        auto_close: bool = True,
        auto_size: bool = True,
        run_async: bool = False,
        result_as: Literal["window", "below", "right"] = "window",
    ) -> ParametricWindow[_W]:
        _, area = self._current_or_new_tab()
        return area.add_parametric_widget(
            widget, callback, title=title, preview=preview, auto_close=auto_close,
            auto_size=auto_size, result_as=result_as, run_async=run_async,
        )  # fmt: skip

    def read_file(
        self,
        file_path: PathOrPaths,
        plugin: str | None = None,
    ) -> SubWindow[_W]:
        """Read local file(s) and open as a new sub-window."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.read_file(file_path, plugin=plugin)

    def read_files(self, file_paths: PathOrPaths):
        """Read multiple files one by one and open as new sub-windows in a same tab."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.read_files(file_paths)

    def read_files_async(self, file_paths: PathOrPaths, plugin: str | None = None):
        """Read multiple files asynchronously and open as new sub-windows."""
        _, tabarea = self._current_or_new_tab()
        return tabarea.read_files_async(file_paths, plugin=plugin)

    def load_session(self, path: str | Path) -> None:
        """Read a session file and update the main window based on the content."""
        from himena.session import update_from_zip, update_from_directory

        fp = Path(path)
        if fp.suffix == ".zip":
            update_from_zip(self, fp)
        elif fp.is_dir():
            update_from_directory(self, fp)
        else:
            raise ValueError(f"Session must be a zip file or a directory, got {fp}.")
        # always plugin=None for reading a session file as a session
        self._recent_session_manager.append_recent_files([(fp, None)])
        self.set_status_tip(f"Session loaded: {fp}", duration=5)

    def save_session(
        self,
        path: str | Path,
        *,
        save_copies: bool = False,
        allow_calculate: Sequence[str] = (),
    ) -> None:
        """Save the current session to a zip file as a stand-along file."""
        from himena.session import dump_zip, dump_directory

        path = Path(path)
        if path.suffix == ".zip":
            dump_zip(
                self, path, save_copies=save_copies, allow_calculate=allow_calculate
            )
        else:
            dump_directory(
                self, path, save_copies=save_copies, allow_calculate=allow_calculate
            )
        self.set_status_tip(f"Session saved to {path}")
        self._recent_session_manager.append_recent_files([(path, None)])

    def clear(self) -> None:
        """Clear all widgets in the main window."""
        self.tabs.clear()
        self.dock_widgets.clear()

    def set_status_tip(self, text: str, duration: float = 10.0) -> None:
        """Set the status tip of the main window.

        This method can be safely called from any thread.

        Parameters
        ----------
        text : str
            Text to show in the status bar.
        duration : float, default 10.0
            Duration (seconds) to show the status tip.
        """
        self._backend_main_window._set_status_tip(text, duration)
        return None

    def show_notification(self, text: str, duration: float = 5.0) -> None:
        """Show a temporary notification in the main window."""
        self._backend_main_window._show_notification(text, duration)

    @overload
    def register_function(
        self,
        func: None = None,
        *,
        menus: str | Sequence[str] = "plugins",
        title: str | None = None,
        types: str | Sequence[str] | None = None,
        enablement: BoolOp | None = None,
        keybindings: Sequence[KeyBindingRule] | None = None,
        command_id: str | None = None,
    ) -> None: ...  # noqa: E501
    @overload
    def register_function(
        self,
        func: _F,
        *,
        menus: str | Sequence[str] = "plugins",
        title: str | None = None,
        types: str | Sequence[str] | None = None,
        enablement: BoolOp | None = None,
        keybindings: Sequence[KeyBindingRule] | None = None,
        command_id: str | None = None,
    ) -> _F: ...  # noqa: E501

    def register_function(
        self,
        func=None,
        *,
        menus="plugins",
        title=None,
        types=None,
        enablement=None,
        keybindings=None,
        command_id=None,
    ):
        """Register a function as a callback in runtime.

        Example
        -------
        ``` python
        @ui.register_function(menus="plugins", title="Test functions)
        def test_function():
            print("test")
        ```

        Parameters
        ----------
        func : callable, optional
            Function to register as an action.
        menus : str or sequence of str, default "plugins"
            Menu(s) to add the action. Submenus are separated by `/`.
        title : str, optional
            Title of the action. Name of the function will be used if not given.
        types: str or sequence of str, optional
            The `type` parameter(s) allowed as the WidgetDataModel. If this parameter
            is given, action will be grayed out if the active window does not satisfy
            the listed types.
        enablement: Expr, optional
            Expression that describes when the action will be enabled. As this argument
            is a generalized version of `types` argument, you cannot use both of them.
        command_id : str, optional
            Command ID. If not given, the function qualname will be used.
        """

        def _inner(f):
            action = _actions.make_action_for_function(
                f,
                menus=menus,
                title=title,
                types=types,
                enablement=enablement,
                keybindings=keybindings,
                command_id=command_id,
            )
            _actions.AppActionRegistry.instance().add_action(action)
            added_menus = _actions.AppActionRegistry.instance().install_to(
                self.model_app, [action]
            )
            self._backend_main_window._rebuild_for_runtime(added_menus)
            return f

        return _inner(func) if func else _inner

    def exec_action(
        self,
        id: str,
        *,
        model_context: WidgetDataModel | None = None,
        window_context: SubWindow | None = None,
        with_params: dict[str, Any] | None = None,
        process_model_output: bool = True,
    ) -> Any:
        """Execute an action by its ID.

        Parameters
        ----------
        id : str
            Action ID.
        model_context : WidgetDataModel, optional
            If given, this model will override the application context for the type
            `WidgetDataModel` before the execution.
        window_context : SubWindow, optional
            If given, this window will override the application context for the type
            `SubWindow` before the execution.
        with_params : dict, optional
            Parameters to pass to the parametric action. These parameters will directly
            be passed to the parametric window created after the action is executed.
        process_model_output : bool, default True
            If True, the output result will be processed by the application context. If
            the command return a `WidgetDataModel` instance, it will be converted to a
            sub-window.
        """
        providers: list[tuple[Any, type]] = []
        if model_context is not None:
            providers.append((model_context, WidgetDataModel, 1000))
        if window_context is not None:
            if isinstance(window_context, SubWindow):
                _window_context = window_context
            else:
                raise TypeError(
                    f"`window_context` must be SubWindow or UUID, got {window_context}"
                )
            providers.append((_window_context, SubWindow, 1000))
            if model_context is None and _window_context.supports_to_model:
                providers.append((_window_context.to_model(), WidgetDataModel, 100))
        # execute the command under the given context
        with (
            self.model_app.injection_store.register(providers=providers),
            self._execute_in_context(
                is_gui=with_params is None, process_model_output=process_model_output
            ),
        ):
            result = self.model_app.commands.execute_command(id).result()
            if with_params is not None:
                if (tab := self.tabs.current()) is not None and len(tab) > 0:
                    param_widget = tab[-1]
                else:  # pragma: no cover
                    raise RuntimeError("Unreachable code.")
                if not isinstance(param_widget, ParametricWindow):
                    if len(with_params) == 0:
                        if isinstance(result, Future):
                            return result.result()  # or appropriate handling
                        return result
                    raise ValueError(
                        f"Parametric widget expected but got {param_widget}."
                    )
                # run the callback with the given parameters synchronously
                result = param_widget._callback_with_params(
                    with_params,
                    force_sync=True,
                    force_close=True,
                )
        return result

    @overload
    def exec_choose_one_dialog(
        self,
        title: str,
        message: str,
        choices: list[tuple[str, _T]],
        how: Literal["buttons", "radiobuttons"] = "buttons",
    ) -> _T | None: ...
    @overload
    def exec_choose_one_dialog(
        self,
        title: str,
        message: str,
        choices: list[str],
        how: Literal["buttons", "radiobuttons"] = "buttons",
    ) -> str | None: ...

    def exec_choose_one_dialog(self, title, message, choices, how="buttons"):
        """Execute a dialog to choose one from the given choices.

        Parameters
        ----------
        title : str
            Window title of the dialog.
        message : str
            HTML Message to show in the dialog.
        choices : list
            List of choices. Each choice can be a string or a tuple of (text, value).
            This method will return the selected value.
        how : str, default "buttons"
            How to show the choices. "buttons" for horizontal buttons, "radiobuttons"
            for vertically arranged radio buttons.
        """
        if res := self._instructions.choose_one_dialog_response:
            return res()
        _choices_normed = []
        for choice in choices:
            if isinstance(choice, str):
                _choices_normed.append((choice, choice))
            else:
                text, value = choice
                _choices_normed.append((text, value))
        return self._backend_main_window._request_choice_dialog(
            title, message, _choices_normed, how=how
        )

    @overload
    def exec_file_dialog(
        self,
        mode: Literal["r", "d", "w"] = "r",
        *,
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: str | Path | None = None,
        group: str | None = None,
    ) -> Path | None: ...
    @overload
    def exec_file_dialog(
        self,
        mode: Literal["rm"],
        *,
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: str | Path | None = None,
        group: str | None = None,
    ) -> list[Path] | None: ...

    def exec_file_dialog(
        self,
        mode: Literal["r", "d", "w", "rm"] = "r",
        *,
        extension_default=None,
        allowed_extensions=None,
        caption=None,
        start_path=None,
        group: str | None = None,
    ):
        """Execute a file dialog to get file path(s)."""
        if mode not in {"r", "d", "w", "rm"}:
            raise ValueError(f"`mode` must be 'r', 'd', 'w' or 'rm', got {mode!r}.")
        if res := self._instructions.file_dialog_response:
            return res()
        if group is None:
            group = mode

        if mode == "w":
            if start_path is None:
                _start_path = self._file_dialog_hist.get_path(group)
            elif Path(start_path).parent != Path("."):
                _start_path = Path(start_path)
            else:  # filename only is given
                _start_path = self._file_dialog_hist.get_path(group, str(start_path))
        else:
            _start_path = Path(start_path or self._file_dialog_hist.get_path(group))
        result = self._backend_main_window._open_file_dialog(
            mode,
            extension_default=extension_default,
            allowed_extensions=allowed_extensions,
            caption=caption,
            start_path=_start_path,
        )
        if result is None:
            return None
        if mode in ["r", "w", "d"]:
            self._file_dialog_hist.update(group, result.parent)
        elif result:
            self._file_dialog_hist.update(group, result[0].parent)
        return result

    def show(self, run: bool = False) -> None:
        """
        Show the main window.

        Parameters
        ----------
        run : bool, default False
            If True, run the application event loop.
        """
        self._backend_main_window.show(run)
        return None

    def close(self) -> None:
        """Close the main window."""
        self._backend_main_window._exit_main_window(confirm=False)
        remove_instance(self.model_app.name, self)
        return None

    @property
    def current_window(self) -> SubWindow[_W] | None:
        """Get the current sub-window."""
        idx_tab = self._backend_main_window._current_tab_index()
        if idx_tab is None:
            return None
        idx_win = self._backend_main_window._current_sub_window_index(idx_tab)
        if idx_win is None:
            return None
        return self.tabs[idx_tab][idx_win]

    @current_window.setter
    def current_window(self, win: SubWindow[_W] | None) -> None:
        """Set the current sub-window."""
        _main = self._backend_main_window
        if win is None:
            _main._set_current_tab_index(None)
            i_tab = _main._current_tab_index()
            if i_tab is not None:
                _main._set_current_sub_window_index(i_tab, None)
        else:
            for i_tab, tab in self.tabs.enumerate():
                for i_win, sub in tab.enumerate():
                    if sub is win:
                        _main._set_current_tab_index(i_tab)
                        _main._set_current_sub_window_index(i_tab, i_win)
                        return None
        return None

    @property
    def current_model(self) -> WidgetDataModel | None:
        """Get the current model of the active sub-window."""
        if sub := self.current_window:
            return sub.to_model()
        return None

    def iter_windows(self) -> Iterator[SubWindow[_W]]:
        """Iterate over all the sub-windows in this main window."""
        for tab in self.tabs:
            yield from tab

    def _provide_file_output(self) -> tuple[WidgetDataModel, SubWindow[_W]]:
        if sub := self.current_window:
            model = sub.to_model()
            return model, sub
        else:
            raise ValueError("No active window.")

    def _tab_activated(self, i: int):
        if i < 0:
            return None
        tab = self.tabs.get(i)
        if tab is not None:
            self.events.tab_activated.emit(tab)
            self._main_window_resized(self.area_size)  # update layout and anchor
        if self._history_tab.get_from_last(1) != i:
            self._history_tab.add(i)
        return None

    def move_window(self, sub: SubWindow[_W], target_index: int) -> None:
        """Move the sub-window to the target tab index."""
        i_tab = i_win = None
        for _i_tab, tab in self.tabs.enumerate():
            for _i_win, win in tab.enumerate():
                if win is sub:
                    i_tab = _i_tab
                    i_win = _i_win
                    break

        if i_tab is None or i_win is None or target_index == i_tab:
            return None
        title = self.tabs[i_tab][i_win].title
        old_rect = self.tabs[i_tab][i_win].rect
        win, widget = self.tabs[i_tab]._pop_no_emit(i_win)
        if target_index < 0:
            self.add_tab()
        self.tabs[target_index].append(win, title)
        win.rect = old_rect
        if layout := win._parent_layout_ref():
            layout.remove(win)
        self.tabs.current_index = i_tab
        return None

    def _window_activated(self):
        back = self._backend_main_window
        back._update_context()
        i_tab = back._current_tab_index()
        if i_tab is None:
            return back._update_control_widget(None)
        tab = self.tabs.get(i_tab)
        if tab is None or len(tab) == 0:
            return back._update_control_widget(None)
        i_win = back._current_sub_window_index(i_tab)
        if i_win is None or len(tab) <= i_win:
            return back._update_control_widget(None)
        win = tab[i_win]
        back._update_control_widget(win.widget)
        _checker.call_widget_activated_callback(win.widget)
        self.events.window_activated.emit(win)

    def _main_window_resized(self, size: Size):
        if tab := self.tabs.current():
            for layout in tab.layouts:
                layout.anchor = layout.anchor.update_for_window_rect(size, layout.rect)
                layout._reanchor(size)
            for win in tab:
                win.anchor = win.anchor.update_for_window_rect(size, win.rect)
                win._reanchor(size)

    @contextmanager
    def _execute_in_context(
        self,
        is_gui: bool = False,
        process_model_output: bool = True,
        unwrap_future: bool = True,
    ):
        with self._global_lock:
            old_inst = self._instructions.model_copy()
            self._instructions = self._instructions.updated(
                gui_execution=is_gui,
                process_model_output=process_model_output,
                unwrap_future=unwrap_future,
            )
            try:
                yield None
            finally:
                self._instructions = old_inst

    def _iter_widget_class(self, model: WidgetDataModel) -> Iterator[type[_W]]:
        """Pick the most suitable widget class for the given model."""
        if model.force_open_with:
            yield import_object(model.force_open_with)
            return
        widget_classes, fallback_class = self._backend_main_window._list_widget_class(
            model.type
        )
        if not widget_classes:
            warnings.warn(
                f"No widget class is registered for model type {model.type!r}.",
                RuntimeWarning,
                stacklevel=2,
            )
            yield fallback_class
            return
        complete_match = [
            (tup.priority, tup.widget_class)
            for tup in widget_classes
            if tup.type == model.type and tup.priority >= 0
        ]
        if complete_match:
            yield from _iter_sorted(complete_match)
        subtype_match = [
            ((tup.type.count("."), tup.priority), tup.widget_class)
            for tup in widget_classes
            if tup.priority >= 0
        ]
        yield from _iter_sorted(subtype_match)

    def _pick_widget(self, model: WidgetDataModel) -> _W:
        """Pick the most suitable widget for the given model."""
        exceptions: list[tuple[Any, Exception]] = []
        for factory in self._iter_widget_class(model):
            try:
                try:
                    widget = factory(self)
                except TypeError:
                    widget = factory()
                widget_id = get_widget_class_id(type(widget))
                reg = _actions.AppActionRegistry.instance()
                if self.model_app.name != "." and (
                    plugin_configs := self.app_profile.plugin_configs.get(widget_id)
                ):
                    params = {}
                    for k, v in plugin_configs.items():
                        params[k] = v["value"]
                    cfgs = reg._plugin_default_configs
                    cfg_type = cfgs[widget_id].config_class
                    # widget should always have `update_configs` in this case
                    widget.update_configs(cfg_type(**params))
                widget.update_model(model)
            except Exception as e:
                exceptions.append((factory, e))
            else:
                break
        else:
            raise ValueError(
                f"Failed to create a widget for {_short_repr(model)}. Errors:\n"
                f"{_format_exceptions(exceptions)}"
            ) from exceptions[-1][1]
        if exceptions:
            raise exceptions[-1][1]

        return widget

    def _on_command_execution(self, id: str, result: Future):
        if exc := result.exception():
            _LOGGER.exception("Command %r failed: %r", id, exc)
            return
        if action := self.model_app.registered_actions.get(id):
            if getattr(action.callback, NO_RECORDING_FIELD, False):
                return None
            self._history_command.add(id)


def _short_repr(obj: Any) -> str:
    obj_repr = repr(obj)
    if len(obj_repr) > 50:
        obj_repr = obj_repr[:50] + "..."
    return obj_repr


def _format_exceptions(exceptions: list[tuple[Any, Exception]]) -> str:
    strs: list[str] = []
    for factory, e in exceptions:
        fname = getattr(factory, "__himena_widget_id__", repr(factory))
        strs.append(f"- {type(e).__name__} in {fname}\n  {e}")
    return "\n".join(strs)


def _iter_sorted(matched: list[tuple[tuple[int, int], type]]) -> Iterator[type]:
    _s = sorted(matched, key=lambda x: x[0], reverse=True)
    yield from (cls for _, cls in _s)



from __future__ import annotations

import numpy as np
from numpy.testing import assert_equal
from himena import WidgetDataModel, StandardType
from himena.standards.model_meta import TableMeta
from himena.testing.subwindow import WidgetTester


def test_accepts_table_like(tester: WidgetTester):
    tester.update_model(
        WidgetDataModel(value=[[1, True], ["b", -5.2]], type=StandardType.TABLE)
    )
    assert_equal(tester.to_model().value, [["1", "True"], ["b", "-5.2"]])
    tester.update_model(
        WidgetDataModel(value={"a": [1, 2], "b": [3, 4]}, type=StandardType.TABLE)
    )
    assert_equal(tester.to_model().value, [["a", "b"], ["1", "3"], ["2", "4"]])
    tester.update_model(
        WidgetDataModel(
            value=np.array([[1, 2], [3, 4]], dtype=np.int64), type=StandardType.TABLE
        )
    )
    assert_equal(tester.to_model().value, [["1", "2"], ["3", "4"]])


def test_current_position(tester: WidgetTester):
    tester.update_model(
        WidgetDataModel(
            value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]], type=StandardType.TABLE
        )
    )
    meta = TableMeta(current_position=[1, 2])
    tester.update_model(
        WidgetDataModel(
            value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            metadata=meta,
            type=StandardType.TABLE,
        )
    )
    meta = _cast_meta(tester.to_model().metadata)
    if meta.current_position != [1, 2]:
        raise AssertionError(f"Expected [1, 2], got {meta.current_position!r}")


def test_selections(tester: WidgetTester):
    tester.update_model(
        WidgetDataModel(
            value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            type=StandardType.TABLE,
        )
    )
    meta = _cast_meta(tester.to_model().metadata)
    if meta.selections != []:
        raise AssertionError(f"Expected [], got {meta.selections}")
    meta = TableMeta(selections=[((1, 2), (0, 2))])
    tester.update_model(
        WidgetDataModel(
            value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            metadata=meta,
            type=StandardType.TABLE,
        )
    )
    meta = _cast_meta(tester.to_model().metadata)
    if meta.selections != [((1, 2), (0, 2))]:
        raise AssertionError(f"Expected [((1, 2), (0, 2))], got {meta.selections}")
    meta = TableMeta(selections=[((1, 2), (0, 2)), ((0, 1), (1, 2))])
    tester.update_model(
        WidgetDataModel(
            value=[[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            metadata=meta,
            type=StandardType.TABLE,
        )
    )
    meta = _cast_meta(tester.to_model().metadata)
    if meta.selections != [((1, 2), (0, 2)), ((0, 1), (1, 2))]:
        raise AssertionError(
            f"Expected [((1, 2), (0, 2)), ((0, 1), (1, 2))], got {meta.selections}"
        )


def _cast_meta(meta) -> TableMeta:
    assert isinstance(meta, TableMeta)
    return meta



from __future__ import annotations
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, Signal


# modified from napari/_qt/widgets/qt_color_swatch.py
class QColorSwatch(QtW.QFrame):
    colorChanged = Signal(QtGui.QColor)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setFrameStyle(QtW.QFrame.Shape.Panel | QtW.QFrame.Shadow.Sunken)
        self._qcolor: QtGui.QColor = QtGui.QColor(255, 255, 255, 255)
        self.colorChanged.connect(self._update_swatch_style)
        self.setMinimumWidth(40)

    @property
    def rgba(self) -> tuple[int, int, int, int]:
        """Get RBGA tuple from QColor."""
        return self._qcolor.getRgb()

    def heightForWidth(self, w: int) -> int:
        return w

    def _update_swatch_style(self, color: QtGui.QColor) -> None:
        rgba = f'rgba({",".join(str(x) for x in color.getRgb())})'
        return self.setStyleSheet("QColorSwatch {background-color: " + rgba + ";}")

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        """Show QColorPopup picker when the user clicks on the swatch."""
        if event.button() == Qt.MouseButton.LeftButton:
            dlg = QtW.QColorDialog(self._qcolor, self)
            dlg.setOptions(QtW.QColorDialog.ColorDialogOption.ShowAlphaChannel)
            ok = dlg.exec_()
            if ok:
                self.setQColor(dlg.selectedColor())

    def qColor(self) -> QtGui.QColor:
        return self._qcolor

    def setQColor(self, color: QtGui.QColor) -> None:
        old_color = self._qcolor
        self._qcolor = color
        if self._qcolor.getRgb() != old_color.getRgb():
            self.colorChanged.emit(color)


class QColorLineEdit(QtW.QLineEdit):
    colorChanged = Signal()

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self._qcolor = QtGui.QColor(255, 255, 255, 255)
        self.editingFinished.connect(self._emit_color_changed)

    def qColor(self) -> QtGui.QColor:
        """Get color as QtGui.QColor object"""
        return self._qcolor

    def setQColor(self, color: QtGui.QColor):
        self._qcolor = color
        text = color.name()
        self.setText(text)

    def _emit_color_changed(self):
        text = self.text()
        try:
            qcolor = QtGui.QColor(text)
        except ValueError:
            self.setQColor(self._qcolor)
        else:
            if self._qcolor.getRgb() != qcolor.getRgb():
                self._qcolor = qcolor
                self.colorChanged.emit()


class QColorEdit(QtW.QWidget):
    colorChanged = Signal(tuple)

    def __init__(self, parent=None):
        super().__init__(parent)
        _layout = QtW.QHBoxLayout(self)
        self._color_swatch = QColorSwatch(self)
        self._line_edit = QColorLineEdit(self)
        _layout.addWidget(self._color_swatch)
        _layout.addWidget(self._line_edit)
        _layout.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(
            QtW.QSizePolicy.Policy.MinimumExpanding, QtW.QSizePolicy.Policy.Fixed
        )
        self._color_swatch.colorChanged.connect(self._on_swatch_changed)
        self._line_edit.colorChanged.connect(self._on_line_edit_edited)
        self.setSizePolicy(QtW.QSizePolicy.Policy.Fixed, QtW.QSizePolicy.Policy.Fixed)
        self.setMinimumHeight(20)

    def color(self) -> QtGui.QColor:
        """Return the current color."""
        return self._color_swatch._qcolor

    def rgba(self) -> tuple[float, float, float, float]:
        return self.color().getRgbF()

    def setColor(self, color: QtGui.QColor | Qt.GlobalColor):
        """Set value as the current color."""
        if isinstance(color, Qt.GlobalColor):
            color = QtGui.QColor(color)
        self._line_edit.setQColor(color)
        self._color_swatch.setQColor(color)

    def setRgba(self, rgba):
        qcolor = QtGui.QColor.fromRgbF(*rgba)
        self.setColor(qcolor)

    def _on_line_edit_edited(self):
        self._line_edit.blockSignals(True)
        qcolor = self._line_edit.qColor()
        self._color_swatch.setQColor(qcolor)
        self._line_edit.blockSignals(False)
        self.colorChanged.emit(qcolor.getRgb())

    def _on_swatch_changed(self, qcolor: QtGui.QColor):
        self._color_swatch.blockSignals(True)
        try:
            self._line_edit.setQColor(qcolor)
        finally:
            self._color_swatch.blockSignals(False)
        self.colorChanged.emit(qcolor.getRgb())



from __future__ import annotations

from pathlib import Path
import logging
from typing import TYPE_CHECKING, Any, Callable, Mapping, Sequence
import weakref
from magicgui import widgets as mgw
from qtpy import QtWidgets as QtW, QtCore
from himena.plugins import validate_protocol, _checker, register_hidden_function
from himena.qt._utils import get_main_window
from himena.types import DropResult, Parametric, Size, WidgetDataModel
from himena.consts import StandardType
from himena._utils import unwrap_lazy_model
from himena.workflow import ProgrammaticMethod
from himena_builtins.qt.widgets._splitter import QSplitterHandle
from himena.qt import drag_command
from himena import _drag
from himena_builtins.qt.widgets._dragarea import QDraggableArea

if TYPE_CHECKING:
    from himena.widgets import MainWindow
    from himena.style import Theme

_LOGGER = logging.getLogger(__name__)
_WIDGET_ROLE = QtCore.Qt.ItemDataRole.UserRole
_MODEL_ROLE = QtCore.Qt.ItemDataRole.UserRole + 1


class QModelStack(QtW.QSplitter):
    """A widget that contains a list of models.

    ## Basic Usage

    - This widget represents a list of models of any types. Widget registered for each
      model will be created and displayed on the right side of this widget.
    - The control widget will be updated according to the selected model.
    - Multiple model stacks can be nested.
    - Items can be either **eager** type or **lazy** type. An eager type item is already
      loaded into the memory. A lazy type item only holds how to load the data, and the
      data loading only happens when the item is clicked. If a model stack is created by
      reading a directory or a file group, its items are **always lazy items**.


    ## Drag and Drop

    - Selected items can be dragged out using the drag indicator.
      - If only one item is selected, the internal model will be dragged.
      - If multiple items are selected, a list of model will be dragged, as a model of
        type `StandardType.MODELS` ("models").
    - If the content is lazy items, they will be loaded into memory. The returned items
      are **always eager items**.

    ## Notes

    - If a directory or a group of files are opened and displayed as a model stack, it
      looks very similar to a file explorer. However, the internal data is not the file
      system itself. Deleting the components does NOT delete the file itself.
    - If the content is modified and one of other items is activated, the old data will
      not be updated if the item is a lazy item, but will be updated if it is an eager
      item. This is a very plausible behavior. A lazy item only holds the function to
      load the data, so that updating the returned value of the function does nothing to
      the function itself. An eager item is already loaded into the memory, so that
      updating the data will directly overwrite the content in the memory.
    """

    __himena_widget_id__ = "builtins:QModelStack"
    __himena_display_name__ = "Built-in Model Stack"

    def __init__(self, ui: MainWindow):
        super().__init__(QtCore.Qt.Orientation.Horizontal)
        self._ui = ui
        left = QtW.QWidget()
        layout = QtW.QVBoxLayout(left)
        layout.setContentsMargins(0, 0, 0, 0)
        self._save_btn = QtW.QPushButton("Save")
        self._save_btn.setToolTip("Save the current model to a file.")
        self._save_btn.clicked.connect(self._save_current)
        self._delete_btn = QtW.QPushButton("Del")
        self._delete_btn.setToolTip(
            "Delete the current item from this list. This action does not delete the "
            "original file."
        )
        self._delete_btn.clicked.connect(self._delete_current)
        btn_layout = QtW.QHBoxLayout()
        btn_layout.setContentsMargins(0, 0, 0, 0)
        btn_layout.addWidget(self._save_btn)
        btn_layout.addWidget(self._delete_btn)

        self._model_list = QModelListWidget(self)
        left.setFixedWidth(160)
        self._widget_stack = QStackedModelWidget(self)
        layout.addWidget(self._model_list)
        layout.addLayout(btn_layout)

        self.addWidget(left)
        self.addWidget(self._widget_stack)
        self._model_list.currentItemChanged.connect(self._current_changed)
        self._last_index: int | None = None
        self.setSizes([160, 320])

        self._control_widget = QtW.QStackedWidget()
        self._is_editable = True

    def createHandle(self):
        return QSplitterHandle(self, "left")

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        value = model.value
        name_model_list: list[tuple[str, WidgetDataModel]] = []
        if isinstance(value, Sequence):
            for each in value:
                if isinstance(each, WidgetDataModel):
                    name_model_list.append((each.title, each))
                elif (
                    isinstance(each, Sequence)
                    and len(each) == 2
                    and isinstance(each[1], WidgetDataModel)
                ):
                    name_model_list.append((str(each[0]), each[1]))
                else:
                    raise TypeError(
                        f"Expected a WidgetDataModel or (name, WidgetDataModel), got "
                        f"{type(each)}"
                    )
        elif isinstance(value, Mapping):
            for key, each in value.items():
                if not isinstance(each, WidgetDataModel):
                    raise TypeError(
                        f"Expected WidgetDataModel as the values, got {type(each)}"
                    )
                name_model_list.append((key, each))
        else:
            raise TypeError(f"Expected Sequence or Mapping, got {type(value)}")

        # clear the stack
        while w := self._widget_stack.currentWidget():
            self._widget_stack.removeWidget(w)

        # add new widgets one by one
        self._model_list.clear()
        for name, model in name_model_list:
            if model.type == StandardType.LAZY:
                item = self._make_lazy_item(name, model)
            else:
                item = self._make_eager_item(name, model)
            self._model_list.addItem(item)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        models: list[WidgetDataModel] = []
        for ith in range(self._model_list.count()):
            item = self._model_list.item(ith)
            model = self._model_list.model_at_index(ith)
            if model is None:
                continue  # this should not happen
            name = item.text()
            model.title = name
            models.append(model)
        return WidgetDataModel(
            value=models,
            type=StandardType.MODELS,
            extension_default=".zip",
        )

    @validate_protocol
    def model_type(self) -> StandardType:
        return StandardType.MODELS

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 480, 380

    @validate_protocol
    def control_widget(self):
        return self._control_widget

    @validate_protocol
    def is_editable(self):
        return self._is_editable

    @validate_protocol
    def set_editable(self, editable: bool):
        self._is_editable = editable
        self._model_list.setEditTriggers(
            QtW.QAbstractItemView.EditTrigger.DoubleClicked
            | QtW.QAbstractItemView.EditTrigger.EditKeyPressed
            if editable
            else QtW.QAbstractItemView.EditTrigger.NoEditTriggers
        )
        self._delete_btn.setEnabled(editable)

    @validate_protocol
    def dropped_callback(self, model: WidgetDataModel):
        if model.type == StandardType.LAZY:
            item = self._make_lazy_item(model.title, model)
        else:
            item = self._make_eager_item(model.title, model)
        self._model_list.addItem(item)

    @validate_protocol
    def widget_resized_callback(self, size_old: Size, size_new: Size):
        widget = self._widget_stack.current_interface()
        if widget:
            lw = self._model_list.width()
            _old = size_old.with_width(max(size_old.width - lw, 10))
            _new = size_new.with_width(max(size_new.width - lw, 10))
            _checker.call_widget_resized_callback(widget, _old, _new)

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        if widget := self._widget_stack.current_interface():
            _checker.call_theme_changed_callback(widget, theme)

    def _make_lazy_item(self, name: str, model: WidgetDataModel):
        """Make a list item that will convert a file into a widget when needed."""
        item = QtW.QListWidgetItem(name)
        item.setFlags(item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)
        item.setData(_MODEL_ROLE, model)
        item.setData(_WIDGET_ROLE, None)
        item.setToolTip(_make_tooltip(name, model))
        return item

    def _make_eager_item(self, name: str, model: WidgetDataModel):
        """Make a list item that will immediately converted intoa widget."""
        item = QtW.QListWidgetItem(name)
        item.setFlags(item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)
        item.setData(_MODEL_ROLE, None)
        item.setToolTip(_make_tooltip(name, model))
        widget = self._model_to_widget(model)
        self._add_widget(item, widget)
        return item

    def _model_to_widget(self, model: WidgetDataModel) -> Any:
        # this may return the interface, not the QWidget!
        return self._ui._pick_widget(model)

    def _add_widget(self, item: QtW.QListWidgetItem, widget: Any):
        interf, native_widget = _split_widget_and_interface(widget)
        self._widget_stack.add_widget(interf, native_widget)

        if hasattr(interf, "control_widget"):
            self._control_widget.addWidget(interf.control_widget())
        else:
            self._control_widget.addWidget(QtW.QWidget())  # empty
        item.setData(_WIDGET_ROLE, interf)
        _checker.call_widget_added_callback(interf)
        _checker.call_theme_changed_callback(interf, self._ui.theme)

    def _update_current_index(self):
        row = self._model_list.currentRow()
        item = self._model_list.item(row)
        if item is None:
            return  # the last item is deleted
        widget = item.data(_WIDGET_ROLE)
        if widget is None:
            model = item.data(_MODEL_ROLE)
            if model is None:
                widget = QtW.QLabel("Not Available")
            else:
                model = _exec_lazy_loading(model)
                widget = self._model_to_widget(model)
            self._add_widget(item, widget)
        _, native_widget = _split_widget_and_interface(widget)
        idx = self._widget_stack.indexOf(native_widget)
        self._widget_stack.setCurrentIndex(idx)
        self._control_widget.setCurrentIndex(idx)
        self._control_widget.currentWidget().setVisible(True)
        self._control_widget.setVisible(True)

    def _current_changed(self):
        if self._last_index is not None:
            item = self._model_list.item(self._last_index)
            if (
                (model := item.data(_MODEL_ROLE))
                and (widget := item.data(_WIDGET_ROLE))
                and not _is_modified(widget)
            ):
                item.setData(_WIDGET_ROLE, None)
                item.setData(_MODEL_ROLE, model)
                _, native_widget = _split_widget_and_interface(widget)
                assert isinstance(native_widget, QtW.QWidget)
                stack_idx = self._widget_stack.indexOf(native_widget)
                self._widget_stack.removeWidget(native_widget)
                native_widget.deleteLater()
                if ctrl_widget := self._control_widget.widget(stack_idx):
                    self._control_widget.removeWidget(ctrl_widget)
                    ctrl_widget.deleteLater()

        self._update_current_index()
        self._last_index = self._model_list.currentRow()

    def _save_current(self):
        if widget := self._widget_stack.current_interface():
            model = widget.to_model()
            assert isinstance(model, WidgetDataModel)
            return self._ui.exec_file_dialog(
                "w",
                extension_default=model.extension_default,
                allowed_extensions=model.extensions,
                caption="Save model to ...",
            )
        return False

    def _delete_current(self):
        ith = self._model_list.currentRow()
        widget = self._widget_stack.widget(ith)
        if _is_modified(widget):
            request = self._ui.exec_choose_one_dialog(
                title="Closing window",
                message="The model has been modified. Do you want to save it?",
                choices=["Yes", "No", "Cancel"],
            )
            if request is None or request == "Cancel":
                return None
            elif request == "Yes" and not self._save_current():
                return None
        self._delete_widget(ith)

    def _delete_widget(self, row: int):
        if widget := self._widget_stack.widget(row):
            self._widget_stack.removeWidget(widget)
        self._model_list.takeItem(row)
        if control := self._control_widget.widget(row):
            self._control_widget.removeWidget(control)


def _is_modified(widget: QtW.QWidget):
    return hasattr(widget, "is_modified") and widget.is_modified()


class QModelListWidget(QtW.QListWidget):
    def __init__(self, parent: QModelStack):
        super().__init__(parent)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setEditTriggers(
            QtW.QAbstractItemView.EditTrigger.DoubleClicked
            | QtW.QAbstractItemView.EditTrigger.EditKeyPressed
        )
        self._is_dragging = False
        self._model_stack_ref = weakref.ref(parent)
        self._hover_drag_indicator = QDraggableArea(self)
        self._hover_drag_indicator.setWindowFlags(
            QtCore.Qt.WindowType.FramelessWindowHint
        )
        self._hover_drag_indicator.setFixedSize(14, 14)
        self._hover_drag_indicator.hide()
        self._hover_drag_indicator.dragged.connect(self._exec_drag)
        self._indicator_index: int = -1
        self.setMouseTracking(True)
        self.setAcceptDrops(True)

    def getter_for_item(
        self, item: QtW.QListWidgetItem
    ) -> Callable[[], WidgetDataModel]:
        return lambda: self.model_for_item(item)

    def model_for_item(self, item: QtW.QListWidgetItem) -> WidgetDataModel:
        model = item.data(_MODEL_ROLE)
        if model is None:  # not a lazy item
            model = item.data(_WIDGET_ROLE).to_model()
        else:
            model = _exec_lazy_loading(model)
        model.title = item.text()
        return model

    def model_at_index(self, row: int) -> WidgetDataModel | None:
        item = self.item(row)
        if item is None:
            return None
        return self.model_for_item(item)

    def mouseMoveEvent(self, e):
        if e.button() == QtCore.Qt.MouseButton.NoButton:
            self._hover_event(e.pos())
        return super().mouseMoveEvent(e)

    def _hover_event(self, pos: QtCore.QPoint):
        index = self.indexAt(pos)
        if index.isValid():
            self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)
            index_rect = self.visualRect(index)
            top_right = index_rect.topRight()
            top_right.setX(top_right.x() - 14)
            top_right.setY(top_right.y() + int(index_rect.height() / 2) - 7)
            self._hover_drag_indicator.move(top_right)
            self._hover_drag_indicator.show()
            self._indicator_index = index.row()
        else:
            self.setCursor(QtCore.Qt.CursorShape.ArrowCursor)
            self._hover_drag_indicator.hide()

    def leaveEvent(self, a0):
        self._hover_drag_indicator.hide()
        return super().leaveEvent(a0)

    # Drag events. If a model is dropped to the list widget, add the model to the list.
    # If selected items are dragged, create a list of models and drag it.
    def dragEnterEvent(self, e):
        if _drag.get_dragging_model():
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        e.acceptProposedAction()

    def dropEvent(self, e):
        if (drag_model := _drag.drop()) and (stack := self._model_stack_ref()):
            model = drag_model.data_model()
            if model.type == StandardType.LAZY:
                item = stack._make_lazy_item(model.title, model)
            else:
                item = stack._make_eager_item(model.title, model)
            stack._model_list.addItem(item)

    def _on_drag(self):
        items = self.selectedItems()
        if item_under_cursor := self.item(self._indicator_index):
            item_under_cursor.setSelected(True)
        _s = "s" if len(items) > 1 else ""
        desc = f"{len(items)} item{_s}"
        if len(items) == 0:
            return None
        if len(items) == 1:
            if item_model := items[0].data(_MODEL_ROLE):
                model_type = item_model.type
            else:
                model_type = None
            return drag_command(
                source=self,
                command_id="builtins:QModelStack:choose-one",
                type=model_type,
                with_params={"index": self.row(items[0])},
                desc=desc,
                exec=False,
            )
        else:
            return drag_command(
                source=self,
                command_id="builtins:QModelStack:select-models",
                type=self._model_stack_ref().model_type(),
                with_params={"index": [self.row(item) for item in items]},
                desc=desc,
                exec=False,
            )

    def _exec_drag(self):
        if drag := self._on_drag():
            drag.exec()


class QStackedModelWidget(QtW.QStackedWidget):
    def __init__(self, parent: QModelStack):
        super().__init__(parent)
        self._model_stack_ref = weakref.ref(parent)
        self._widget_to_interface_map = weakref.WeakKeyDictionary[QtW.QWidget, Any]()
        self.setAcceptDrops(True)

    def add_widget(self, interface: Any, widget: QtW.QWidget):
        self.addWidget(widget)
        self._widget_to_interface_map[widget] = interface

    def current_interface(self) -> Any | None:
        if widget := self.currentWidget():
            return self._widget_to_interface_map.get(widget)
        return None

    # Drag events. If a model is dropped to the stacked widget, look for the widget
    # interface and check if it implements the drop event. If so, call the drop event.
    def dragEnterEvent(self, e):
        if (
            (model := _drag.get_dragging_model())
            and (interf := self.current_interface())
            and (model.widget_accepts_me(interf))
        ):
            e.accept()
        else:
            e.ignore()

    def dragMoveEvent(self, e):
        e.acceptProposedAction()

    def dropEvent(self, e):
        self._drop_event()

    def _drop_event(self):
        if (
            (drag_model := _drag.drop())
            and (widget := self.current_interface())
            and hasattr(widget, "dropped_callback")
        ):
            model = drag_model.data_model()
            drop_result = widget.dropped_callback(model)
            if drop_result is None:
                drop_result = DropResult()
            if command_id := drop_result.command_id:
                ui = get_main_window(self)
                model: WidgetDataModel = widget.to_model()
                model.workflow = ProgrammaticMethod().construct_workflow()
                out = ui.exec_action(
                    command_id,
                    with_params=drop_result.with_params,
                    model_context=model,
                    process_model_output=False,
                )
                widget.update_model(out)


@register_hidden_function(command_id="builtins:QModelStack:choose-one")
def _choose_one(model: WidgetDataModel) -> Parametric:
    """Choose one model from the list."""

    def run(index: int) -> WidgetDataModel:
        return model.value[index]

    return run


@register_hidden_function(command_id="builtins:QModelStack:select-models")
def _select_models(model: WidgetDataModel) -> Parametric:
    """Select models from the list."""

    def run(index: Sequence[int]) -> WidgetDataModel:
        return WidgetDataModel(
            value=[model.value[i] for i in index],
            type=StandardType.MODELS,
            title=model.title,
        )

    return run


def _exec_lazy_loading(model: WidgetDataModel) -> WidgetDataModel:
    """Run the pending lazy loading."""
    _LOGGER.info("Lazy loading of: %r", model)
    path = model.source
    model = unwrap_lazy_model(model)
    # determine the title
    if isinstance(path, Path):
        model.title = path.name
    else:
        model.title = "File Group"
    return model


def _split_widget_and_interface(widget) -> tuple[Any, QtW.QWidget]:
    if hasattr(widget, "native_widget"):
        interf = widget
        native_widget = interf.native_widget()
    elif isinstance(widget, mgw.Widget):
        interf = widget
        native_widget = interf.native
    elif isinstance(widget, QtW.QWidget):
        interf = native_widget = widget
    else:
        raise TypeError(f"Expected a widget or an interface, got {type(widget)}")
    return interf, native_widget


def _make_tooltip(name: str, model: WidgetDataModel) -> str:
    attrs: list[str] = [f"<b>Title</b>: {name}"]
    attrs.append(f"<b>Type</b>: {model.type}")
    return "<br>".join(attrs)



from __future__ import annotations
import json

import numpy as np
from himena.standards.roi._base import RoiModel
from himena.utils.ndobject import NDObjectCollection


class RoiListModel(NDObjectCollection[RoiModel]):
    """List of ROIs, with useful methods."""

    def model_dump_typed(self) -> dict:
        return {
            "rois": [roi.model_dump_typed() for roi in self],
            "indices": self.indices.tolist() if self.indices is not None else None,
            "axis_names": self.axis_names,
        }

    @classmethod
    def construct(cls, dict_: dict) -> RoiListModel:
        """Construct an instance from a dictionary."""
        rois = []
        for roi_dict in dict_["rois"]:
            if not isinstance(roi_dict, dict):
                raise ValueError(f"Expected a dictionary for 'rois', got: {roi_dict!r}")
            roi_type = roi_dict.pop("type")
            roi = RoiModel.construct(roi_type, roi_dict)
            rois.append(roi)
        return cls(
            items=rois,
            indices=np.array(dict_["indices"], dtype=np.int32),
            axis_names=dict_["axis_names"],
        )

    @classmethod
    def model_validate_json(cls, text: str) -> RoiListModel:
        """Validate the json string and return an instance."""
        js = json.loads(text)
        return cls.construct(js)



from himena.plugins import (
    register_dock_widget_action,
    add_default_status_tip,
)
from himena.consts import MenuId
from himena_builtins.qt.favorites._config import FavoriteCommandsConfig

add_default_status_tip(
    short="favorites",
    long="Ctrl+Shift+F to edit and run favorite commands",
)


@register_dock_widget_action(
    title="Favorite Commands",
    menus=[MenuId.TOOLS_DOCK, MenuId.CORNER],
    area="right",
    keybindings=["Ctrl+Shift+F"],
    singleton=True,
    command_id="builtins:favorite-commands",
    plugin_configs=FavoriteCommandsConfig(),
    icon="mdi:favorite-circle",
)
def install_favorite_commands(ui):
    """Show the favorite commands."""
    from himena_builtins.qt.favorites._widget import QFavoriteCommands

    return QFavoriteCommands(ui)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from app_model.types import Action
from logging import getLogger
import json
from himena.consts import MenuId, ActionCategory, ActionGroup, NO_RECORDING_FIELD
from himena.profile import data_dir
from datetime import datetime

if TYPE_CHECKING:
    from himena._app_model import HimenaApplication
    from himena.widgets._main_window import MainWindow

    _PathInput = Path | list[Path]

_LOGGER = getLogger(__name__)


class OpenRecentFunction:
    """Function that open a recent file using the specified plugin."""

    def __init__(self, file: _PathInput, plugin: str | None = None):
        self._file = file
        self._plugin = plugin
        setattr(self, NO_RECORDING_FIELD, True)  # don't record "open" command

    def __call__(self, ui: MainWindow):
        _LOGGER.debug("Calling OpenRecentFunction for %s", self._file)
        ui.read_file(self._file, plugin=self._plugin)

    def to_str(self) -> str:
        return f"Open {self._file.as_posix()}"


class OpenSessionFunction:
    def __init__(self, file: _PathInput):
        self._file = file

    def __call__(self, ui: MainWindow):
        ui.load_session(self._file)

    def to_str(self) -> str:
        return f"Load session {self._file.as_posix()}"


class RecentFileManager:
    _MENU_UPDATED: set[str] = set()

    def __init__(
        self,
        app: HimenaApplication,
        menu_id: MenuId = MenuId.FILE_RECENT,
        file_name: str = "recent.json",
        group: str = ActionGroup.RECENT_FILE,
        n_history: int = 60,
        n_history_menu: int = 8,
    ):
        self._disposer = lambda: None
        self._app = app
        self._menu_id = menu_id
        self._file_name = file_name
        self._group = group
        self._n_history = n_history
        self._n_history_menu = n_history_menu

    def update_menu(self):
        """Update the menu for the recent file list."""
        if self._app.name in self._MENU_UPDATED:
            return None
        file_args = self._list_args_for_recent()[::-1]
        if len(file_args) == 0:
            return None
        actions = [
            self.action_for_file(path, plugin, in_menu=i < self._n_history_menu)
            for i, (path, plugin) in enumerate(file_args)
        ]
        self._disposer()
        self._disposer = self._app.register_actions(actions)
        for action in actions:
            self._app._dynamic_command_ids.add(action.id)
        _LOGGER.debug("Recent files updated: %r", [p for p, _ in file_args[:3]])
        self._app.menus.menus_changed.emit({self._menu_id})
        self.__class__._MENU_UPDATED.add(self._app.name)
        return None

    @classmethod
    def default(cls, app: HimenaApplication) -> RecentFileManager:
        return cls(app)

    def _list_args_for_recent(self) -> list[tuple[_PathInput, str | None]]:
        """List the recent files (older first)."""

        _path = data_dir() / self._file_name
        if not _path.exists():
            return []
        with open(_path) as f:
            js = json.load(f)
        if not isinstance(js, list):
            return []
        out: list[tuple[_PathInput, str | None]] = []
        for each in js:
            if not isinstance(each, dict):
                continue
            if "type" not in each:
                continue
            if each["type"] == "group":
                out.append(([Path(p) for p in each["path"]], each.get("plugin")))
            elif each["type"] == "file":
                out.append((Path(each["path"]), each.get("plugin")))
            elif each["type"] == "folder":
                out.append((Path(each["path"]), each.get("plugin")))
        return out

    def append_recent_files(
        self,
        inputs: list[tuple[_PathInput, str | None]],
    ) -> None:
        """Append file(s) with plugin to the user history (duplication OK)."""
        _path = data_dir() / self._file_name
        if _path.exists():
            with open(_path) as f:
                all_info = json.load(f)
            if not isinstance(all_info, list):
                all_info = []
        else:
            all_info = []
        existing_paths = [each["path"] for each in all_info]
        to_remove: list[int] = []
        now = datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        for each_path_input, plugin in inputs:
            if plugin == "himena_builtins.io.read_as_unknown":
                # This plugin should not be used again from the open recent menu.
                plugin = None
            each = _norm_path_input(each_path_input)
            if each in existing_paths:
                to_remove.append(existing_paths.index(each))
            if isinstance(each, list):
                all_info.append(
                    {"type": "group", "path": each, "plugin": plugin, "time": now}
                )
            elif Path(each).is_file():
                all_info.append(
                    {"type": "file", "path": each, "plugin": plugin, "time": now}
                )
            else:
                all_info.append(
                    {"type": "folder", "path": each, "plugin": plugin, "time": now}
                )
        for i in sorted(to_remove, reverse=True):
            all_info.pop(i)
        if len(all_info) > self._n_history:
            all_info = all_info[-self._n_history :]
        _path.write_text(json.dumps(all_info, indent=2))

        # TODO: Don't use Qt!
        from superqt.utils import ensure_main_thread

        ensure_main_thread(self.update_menu)()
        return None

    def action_for_file(
        self,
        file: _PathInput,
        plugin: str | None = None,
        in_menu: bool = True,
    ) -> Action:
        """Make an Action for opening a file."""
        id, title = self.id_title_for_file(file)
        if in_menu:
            menus = [
                {"id": MenuId.RECENT_ALL},
                {"id": self._menu_id, "group": self._group},
            ]
        else:
            menus = [{"id": MenuId.RECENT_ALL}]
        return Action(
            id=id,
            title=title,
            callback=self.to_callback(file, plugin),
            menus=menus,
            category=ActionCategory.OPEN_RECENT,
            palette=False,
        )

    def to_callback(self, file, plugin: str | None = None):
        return OpenRecentFunction(file, plugin)

    def id_title_for_file(self, file: _PathInput) -> tuple[str, str]:
        """Return ID and title for the file."""
        if isinstance(file, Path):
            title = _title_for_file(file)
            id = f"open-{title}"
        else:
            name = ";".join([f.name for f in file])
            id = f"open-{name}"
            title = f"{len(file)} files such as {file[0]}"
        return id, title


class RecentSessionManager(RecentFileManager):
    _MENU_UPDATED: set[str] = set()

    @classmethod
    def default(cls, app: HimenaApplication) -> RecentSessionManager:
        return cls(
            app,
            file_name="recent_sessions.json",
            group=ActionGroup.RECENT_SESSION,
            n_history=20,
            n_history_menu=3,
        )

    def to_callback(self, file, plugin: str | None = None):
        return OpenSessionFunction(file)

    def id_title_for_file(self, file: Path) -> tuple[str, str]:
        """Return the ID for the file."""
        fp = _title_for_file(file)
        id = f"load-session-{fp}"
        title = f"{fp} [Session]"
        return id, title


def _norm_path_input(each: _PathInput) -> str | list[str]:
    if isinstance(each, list):
        return [p.as_posix() for p in each]
    else:
        return each.as_posix()


def _title_for_file(file: Path) -> str:
    home = Path.home()
    if file.is_relative_to(home):
        title = ("~" / file.relative_to(home)).as_posix()
    else:
        title = file.as_posix()
    return title



from __future__ import annotations
from typing import Any

from qtpy import QtWidgets as QtW, QtCore, QtGui
from himena.consts import StandardType, ParametricWidgetProtocolNames as PWPN
from himena.types import ModelTrack, WidgetDataModel
from himena.plugins import validate_protocol
from magicgui import widgets as mgw


class QParametricWidget(QtW.QWidget):
    """QWidget that contain a magicgui Container and a button to run functions."""

    __himena_display_name__ = "Parametric Widget"

    param_changed = QtCore.Signal()

    def __init__(self, central: QtW.QWidget | mgw.Widget) -> None:
        super().__init__()
        self._scroll_area = QtW.QScrollArea()
        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        self._scroll_area.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self._call_btn = QtW.QPushButton("Run", self)
        self._central_widget = central
        layout_h = QtW.QHBoxLayout(self)
        layout_h.setContentsMargins(0, 0, 0, 0)
        layout_h.addWidget(self._scroll_area)
        area_widget = QtW.QWidget()
        self._scroll_area.setWidget(area_widget)
        layout_v = QtW.QVBoxLayout(area_widget)
        layout_v.setContentsMargins(4, 4, 4, 4)
        layout_v.setSpacing(2)
        if isinstance(central, mgw.Widget):
            if not isinstance(central.native, QtW.QWidget):
                raise ValueError(f"Expected a QWidget, got {central}")
            self._central_qwidget = central.native
        else:
            self._central_qwidget = central
        layout_v.addWidget(self._central_qwidget)
        layout_v.addWidget(self._call_btn)
        if connector := getattr(central, PWPN.CONNECT_CHANGED_SIGNAL, None):
            connector(self._on_param_changed)

        if tracker := ModelTrack.get(central):
            tracker.set(self)
        self._result_widget: QtW.QWidget | None = None

        self._control = QtW.QWidget()
        control_layout = QtW.QHBoxLayout(self._control)
        control_layout.setContentsMargins(0, 0, 0, 0)
        self._central_widget_size_hint = self._central_qwidget.sizeHint()
        if isinstance(central, mgw.Container):
            # find the proper size hint
            if len(central) > 0:
                min_height = sum(max(each.min_height, 28) for each in central) + 2 * (
                    len(central) - 1
                )
                min_width = max(max(each.min_width, 80) for each in central) + 6
            else:
                min_height = 20
                min_width = 60
            self._central_widget_size_hint.setHeight(
                max(self._central_widget_size_hint.height(), min_height)
            )
            self._central_widget_size_hint.setWidth(
                max(self._central_widget_size_hint.width(), min_width)
            )
        self._layout_v = layout_v
        self._layout_h = layout_h
        self._result_widget_layout: QtW.QLayout | None = None
        self.setMinimumWidth(100)

    def get_params(self) -> dict[str, Any]:
        return getattr(self._central_widget, PWPN.GET_PARAMS)()

    @validate_protocol
    def to_model(self) -> WidgetDataModel[dict[str, Any]]:
        params = self.get_params()
        return WidgetDataModel(value=params, type=StandardType.DICT)

    @validate_protocol
    def model_type(self: QtW.QWidget) -> str:
        return StandardType.DICT

    @validate_protocol
    def size_hint(self) -> tuple[int, int] | None:
        mysize = self._base_size_hint()
        if self._result_widget is None:
            return mysize.width(), min(mysize.height(), 400)
        if hasattr(self._result_widget, "size_hint") and (
            size := self._result_widget.size_hint()
        ):
            w0, h0 = size
        else:
            hint = self._result_widget.sizeHint()
            w0, h0 = hint.width(), hint.height()
        mysize.setHeight(mysize.height() + h0)
        mysize.setWidth(max(mysize.width(), w0))
        return mysize.width(), min(mysize.height(), 400)

    @validate_protocol
    def control_widget(self) -> QtW.QWidget:
        return self._control

    def keyPressEvent(self, a0: QtGui.QKeyEvent | None) -> None:
        if a0 and a0.key() == QtCore.Qt.Key.Key_Return:
            self._call_btn.click()
        return super().keyPressEvent(a0)

    def _base_size_hint(self) -> QtCore.QSize:
        mysize = QtCore.QSize(self._central_widget_size_hint)
        mysize.setWidth(max(mysize.width() + 10, 80))  # content margins
        mysize.setHeight(mysize.height() + 36)  # content margins and scroll area
        if self._call_btn.isVisible():
            mysize.setHeight(mysize.height() + 24)  # button height
        return mysize

    def setFocus(self) -> None:
        if (
            isinstance(self._central_widget, mgw.Container)
            and len(self._central_widget) > 0
        ):
            # focus the first input
            return self._central_widget[0].native.setFocus()
        else:
            return super().setFocus()

    def _on_param_changed(self) -> None:
        self.param_changed.emit()

    def is_preview_enabled(self) -> bool:
        if isfunc := getattr(self._central_widget, PWPN.IS_PREVIEW_ENABLED, None):
            return isfunc()
        return False

    def add_widget_below(self, widget: QtW.QWidget) -> None:
        self._layout_v.addWidget(widget)
        self._result_widget = widget
        self._result_widget_layout = self._layout_v
        if hasattr(widget, "control_widget"):
            self._control.layout().addWidget(widget.control_widget())

    def add_widget_right(self, widget: QtW.QWidget) -> None:
        self._layout_h.addWidget(widget)
        self._result_widget = widget
        self._result_widget_layout = self._layout_h
        if hasattr(widget, "control_widget"):
            self._control.layout().addWidget(widget.control_widget())

    def remove_result_widget(self) -> None:
        if self._result_widget is not None:
            if self._result_widget_layout is not None:
                self._result_widget_layout.removeWidget(self._result_widget)
            self._result_widget = None
            self._result_widget_layout = None
        if self._control.layout().count() > 0:
            self._control.layout().itemAt(0).widget().setParent(None)

    def set_busy(self, busy: bool):
        if busy:
            self._call_btn.setText("Running ...")
            self._call_btn.setEnabled(False)
        else:
            self._call_btn.setText("Run")
            self._call_btn.setEnabled(True)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, NamedTuple
from himena.consts import StandardType
import numpy as np

if TYPE_CHECKING:
    from himena.types import WidgetDataModel
    from himena.widgets import SubWindow
    from himena.qt.magicgui import SelectionEdit

# Single 2D selection in the form of ((row start, row stop), (col start, col stop))
# We should avoid using slice because it is not serializable.
SelectionType = tuple[tuple[int, int], tuple[int, int]]


class NamedArray(NamedTuple):
    name: str | None
    array: np.ndarray


def model_to_xy_arrays(
    model: WidgetDataModel,
    x: SelectionType | None,
    y: SelectionType | None,
    *,
    allow_empty_x: bool = True,
    allow_multiple_y: bool = True,
    same_size: bool = True,
) -> tuple[NamedArray, list[NamedArray]]:
    """Extract X, Y values from a table-like model.

    How selection works is just like when you plot data in Excel. This function supports
    several types of selections.

    1. Value-only column selection.
       ```
       +++++ +++++
       + 1 + + 3 +
       + 2 + + 6 +
       + 5 + + 9 +
       +++++ +++++
       ```
    2. Selection with names.
       ```
       +++++ +++++
       + x + + y +
       + 1 + + 3 +
       + 2 + + 6 +
       + 5 + + 9 +
       +++++ +++++
       ```
    3. Multiple Y values.
       ```
       +++++ ++++++++
       + x + + y, z +
       + 1 + + 3, 1 +
       + 2 + + 6, 4 +
       + 5 + + 9, 2 +
       +++++ ++++++++
       ```
    """
    from himena.data_wrappers import wrap_dataframe, wrap_array
    from himena.standards.model_meta import DictMeta, ArrayMeta

    if x is None and not allow_empty_x:
        raise ValueError("The x value must be given.")
    if y is None:
        raise ValueError("The y value must be given.")
    if model.is_subtype_of(StandardType.TABLE):
        x_out, ys = table_to_xy_arrays(
            model.value, x, y, allow_multiple_y=allow_multiple_y, same_size=same_size
        )
    elif model.is_subtype_of(StandardType.DATAFRAME):
        df = wrap_dataframe(model.value)
        column_names = df.column_names()[y[1][0] : y[1][1]]
        rows = slice(y[0][0], y[0][1])
        ys = [
            NamedArray(cname, df.column_to_array(cname)[rows]) for cname in column_names
        ]
        if x is None:
            xarr = np.arange(ys[0][1].size)
            xlabel = None
        else:
            column_names_x = df.column_names()[x[1][0] : x[1][1]]
            if len(column_names_x) != 1:
                raise ValueError("x must not be more than one column.")
            xarr = df.column_to_array(column_names_x[0])
            xlabel = column_names_x[0]
        x_out = NamedArray(xlabel, xarr)
    elif model.is_subtype_of(StandardType.EXCEL):
        if not isinstance(meta := model.metadata, DictMeta):
            raise ValueError(f"Must be a DictMeta, got {type(model.metadata)!r}")
        table = model.value[meta.current_tab]
        x_out, ys = table_to_xy_arrays(
            table, x, y, allow_multiple_y=allow_multiple_y, same_size=same_size
        )
    elif model.is_subtype_of(StandardType.ARRAY):
        if not isinstance(meta := model.metadata, ArrayMeta):
            raise ValueError(f"Must be an ArrayMeta, got {type(model.metadata)!r}")
        if meta.current_indices is None:
            sl = ()
        else:
            sl = tuple(
                slice(None) if ind is None else ind for ind in meta.current_indices
            )
        arr = wrap_array(model.value).get_slice(sl)
        x_out, ys = table_to_xy_arrays(
            arr, x, y, allow_multiple_y=allow_multiple_y, same_size=same_size
        )
    else:
        raise ValueError(f"Table-like data expected, but got model type {model.type!r}")
    return x_out, ys


def model_to_vals_arrays(
    model: WidgetDataModel,
    ys: list[SelectionType],
    *,
    same_size: bool = True,
) -> list[NamedArray]:
    values: list[NamedArray] = []
    for yn in ys:
        _, y_out = model_to_xy_arrays(
            model, None, yn, allow_multiple_y=False, same_size=same_size
        )
        values.append(y_out[0])
    if same_size and len({a.array.size for a in values}) != 1:
        raise ValueError("Selection sizes not consistent.")
    return values


def model_to_col_val_arrays(
    model: WidgetDataModel,
    col: SelectionType,
    val: SelectionType,
) -> tuple[NamedArray, NamedArray]:
    """Extract a categorical column and a value column from a table-like model.

    Very similar to `model_to_xy_arrays`, but this function does not require a numerical
    column for the `col` input.
    """
    from himena.data_wrappers import wrap_dataframe
    from himena.standards.model_meta import DictMeta

    if model.is_subtype_of(StandardType.TABLE):
        x_out, y_out = table_to_col_val_arrays(model.value, col, val)
    elif model.is_subtype_of(StandardType.DATAFRAME):
        df = wrap_dataframe(model.value)
        i_col = _to_single_column_slice(col)
        i_val = _to_single_column_slice(val)
        column_names = df.column_names()
        x_out = NamedArray(column_names[i_col], df[column_names[i_col]])
        y_out = NamedArray(column_names[i_val], df[column_names[i_val]])
    elif model.is_subtype_of(StandardType.EXCEL):
        if not isinstance(meta := model.metadata, DictMeta):
            raise ValueError("Must be a DictMeta")
        table = model.value[meta.current_tab]
        x_out, y_out = table_to_col_val_arrays(table, col, val)
    else:
        raise ValueError(f"Table-like data expected, but got model type {model.type!r}")
    return x_out, y_out


def table_to_xy_arrays(
    value: np.ndarray,
    x: SelectionType | None,
    y: SelectionType,
    *,
    allow_empty_x: bool = True,
    allow_multiple_y: bool = True,
    same_size: bool = True,
) -> tuple[NamedArray, list[tuple[NamedArray]]]:
    if x is None and not allow_empty_x:
        raise ValueError("The x value must be given.")
    ysl = slice(y[0][0], y[0][1]), slice(y[1][0], y[1][1])
    parser = TableValueParser.from_array(value[ysl])
    if x is None:
        xarr = np.arange(parser.n_samples, dtype=np.float64)
        xlabel = None
    else:
        xsl = slice(x[0][0], x[0][1]), slice(x[1][0], x[1][1])
        xlabel, xarr = parser.norm_x_value(value[xsl], same_size=same_size)
    if not allow_multiple_y and parser.n_components > 1:
        raise ValueError("Multiple Y values are not allowed.")
    return NamedArray(xlabel, xarr), parser.named_arrays


def table_to_col_val_arrays(
    value: np.ndarray,
    col: SelectionType,
    val: SelectionType,
) -> tuple[NamedArray, NamedArray]:
    col_sl = slice(col[0][0], col[0][1]), slice(col[1][0], col[1][1])
    val_sl = slice(val[0][0], val[0][1]), slice(val[1][0], val[1][1])
    parser = TableValueParser.from_array(value[val_sl])
    if parser.n_components != 1:
        raise ValueError("Multiple Y values are not allowed.")
    col_arr = parser.norm_col_value(value[col_sl])
    return col_arr, parser.named_arrays[0]


class TableValueParser:
    def __init__(
        self,
        label_and_values: list[NamedArray],
        is_column_vector: bool = True,
    ):
        self._label_and_values = label_and_values
        self._is_column_vector = is_column_vector

    @property
    def named_arrays(self) -> list[NamedArray]:
        return self._label_and_values

    @classmethod
    def from_columns(cls, value: np.ndarray) -> TableValueParser:
        nr, nc = value.shape
        if nr == 1:
            return cls([NamedArray(None, as_f64(value[:, i])) for i in range(nc)])
        try:
            as_f64(value[0, :])  # try to cast to float
        except ValueError:
            # The first row is not numerical. Use it as labels.
            return cls(
                [NamedArray(str(value[0, i]), as_f64(value[1:, i])) for i in range(nc)]
            )
        else:
            return cls([NamedArray(None, as_f64(value[:, i])) for i in range(nc)])

    @classmethod
    def from_rows(cls, value: np.ndarray) -> TableValueParser:
        self = cls.from_columns(value.T)
        self._is_column_vector = False
        return self

    @classmethod
    def from_array(cls, value: np.ndarray) -> TableValueParser:
        try:
            return cls.from_columns(value)
        except ValueError:
            return cls.from_rows(value)

    @property
    def n_components(self) -> int:
        return len(self._label_and_values)

    @property
    def n_samples(self) -> int:
        return self._label_and_values[0][1].size

    def norm_x_value(self, arr: np.ndarray, same_size: bool = True) -> NamedArray:
        # check if the first value is a label
        if self._is_column_vector and arr.shape[1] != 1:
            raise ValueError("The X values must be a 1D column vector.")
        if not self._is_column_vector and arr.shape[0] != 1:
            raise ValueError("The X values must be a 1D row vector.")
        arr = arr.ravel()
        try:
            arr[:1].astype(np.float64)
        except ValueError:
            label, arr_number = str(arr[0]), arr[1:].astype(np.float64)
        else:
            label, arr_number = None, arr.astype(np.float64)
        if same_size and arr_number.size != self.n_samples:
            raise ValueError("The number of X values must be the same as the Y values.")
        return NamedArray(label, arr_number)

    def norm_col_value(self, arr: np.ndarray) -> NamedArray:
        # check if the first value is a label
        if self._is_column_vector and arr.shape[1] != 1:
            raise ValueError("The X values must be a 1D column vector.")
        if not self._is_column_vector and arr.shape[0] != 1:
            raise ValueError("The X values must be a 1D row vector.")
        arr = arr.ravel()
        if arr.size == self.n_samples:
            label, arr_out = None, arr
        else:
            label, arr_out = str(arr[0]), arr[1:]
        return NamedArray(label, arr_out)


TABLE_LIKE_TYPES = [
    StandardType.TABLE,
    StandardType.DATAFRAME,
    StandardType.ARRAY,
    StandardType.EXCEL,
]


def range_getter(
    ref: SubWindow | str,
) -> Callable[..., tuple[SelectionType, SelectionType]]:
    """The getter function for SelectionEdit"""
    from himena.standards.model_meta import TableMeta, ArrayMeta
    from magicgui.widgets.bases import ContainerWidget

    def _getter(widget: SelectionEdit):
        if isinstance(ref, str):
            if not isinstance(fgui := widget.parent, ContainerWidget):
                raise ValueError(
                    f"Parent of a selection edit must be a ContainerWidget, but was {type(fgui)}"
                )
            for child in fgui:
                if child.name == ref:
                    model = child.value.to_model()
                    break
        else:
            model = ref.to_model()
        if model.type not in TABLE_LIKE_TYPES:
            raise ValueError(f"Cannot plot model of type {model.type!r}")
        if not isinstance(meta := model.metadata, (TableMeta, ArrayMeta)):
            raise ValueError("Excel must have TableMeta as the additional data.")
        if len(meta.selections) == 0:
            raise ValueError(f"No selection found in window {model.title!r}")
        elif len(meta.selections) > 1:
            raise ValueError(f"More than one selection found in window {model.title!r}")
        sel = meta.selections[0]
        return sel

    return _getter


def table_selection_gui_option(
    ref: SubWindow | str,
    default: SelectionType | None = None,
) -> dict:
    """GUI option used for a parameter of a single table selection.

    This function is always used with `configure_gui()`. If a parameter `x` is a table
    selection type that takes a ((int, int), (int, int)) tuple, then the GUI option can
    be easily configured by this function.
    If the table widget window `win` is known, use the expression

    ```python
    @configure_gui(x=table_selection_gui_option(win, default))
    def inner_function(x: SelectionType | None): ...
    ```

    If the table widget is to be  determined by another parameter named "table"

    ```python
    @configure_gui(
        table={"types": TABLE_LIKE_TYPES}
        x=table_selection_gui_option("table", default)
    )
    def inner_function(table: SubWindow, x: SelectionType | None): ...
    ```

    For many cases, the default value can be nicely determined by the `auto_select()`
    function.
    """
    from himena.qt.magicgui import SelectionEdit

    return {"widget_type": SelectionEdit, "getter": range_getter(ref), "value": default}


def auto_select(model: WidgetDataModel, num: int) -> list[None | SelectionType]:
    """Automatically select a number of columns from a table-like model.

    This function will select the columns from left to right by default, but will first
    select the already selected columns if any.
    """
    from himena.data_wrappers import wrap_dataframe
    from himena.standards.model_meta import TableMeta, ArrayMeta, DictMeta

    selections: list[tuple[tuple[int, int], tuple[int, int]]] = []
    val = model.value
    if model.is_subtype_of(StandardType.TABLE):
        if not isinstance(val, np.ndarray):
            raise ValueError(f"Table must be a numpy array, got {type(val)}")
        shape = val.shape
        if isinstance(meta := model.metadata, TableMeta):
            selections = meta.selections
    elif model.is_subtype_of(StandardType.DATAFRAME):
        df = wrap_dataframe(val)
        shape = df.shape
        if isinstance(meta := model.metadata, TableMeta):
            selections = meta.selections
    elif model.is_subtype_of(StandardType.EXCEL):
        if not isinstance(meta := model.metadata, DictMeta):
            raise ValueError(f"Expected an DictMeta, got {type(meta)}")
        table = val[meta.current_tab]
        if not isinstance(table, np.ndarray):
            raise ValueError(f"Table must be a numpy array, got {type(table)}")
        shape = table.shape
        selections = meta.child_meta[meta.current_tab].selections
    elif model.is_subtype_of(StandardType.ARRAY):
        if not isinstance(val, np.ndarray):
            raise ValueError(f"Array must be a numpy array, got {type(val)}")
        if isinstance(meta := model.metadata, ArrayMeta):
            selections = meta.selections
        shape = val.shape
    else:
        raise ValueError(f"Table-like data expected, but got model type {model.type!r}")
    ncols = shape[1]
    if num == len(selections):
        return selections
    if ncols == 0:
        raise ValueError("The table must have at least one column.")
    elif ncols < num:
        out = [None] * num
        for i in range(ncols):
            out[i + num - ncols] = ((0, shape[0]), (i, i + 1))
        return out
    else:
        return [((0, shape[0]), (i, i + 1)) for i in range(num)]


def _to_single_column_slice(val: SelectionType) -> int:
    _, csl = val
    if csl[1] - csl[0] != 1:
        raise ValueError("Only single column selection is allowed")
    return csl[0]


def as_f64(arr: np.ndarray):
    if arr.dtype.kind == "T":
        return np.where(arr == "", "nan", arr).astype(np.float64)
    return arr.astype(np.float64)



from __future__ import annotations

from qtpy import QtWidgets as QtW


class QControlStack(QtW.QStackedWidget):
    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        # window content -> control widget
        self._empty_widget = QtW.QWidget()
        self._widget_map: dict[QtW.QWidget, QtW.QWidget] = {}
        self.addWidget(self._empty_widget)

    def add_control_widget(
        self,
        widget: QtW.QWidget,
        control: QtW.QWidget,
    ) -> None:
        self.addWidget(control)
        self._widget_map[widget] = control
        self.setCurrentWidget(control)

    def remove_control_widget(self, widget: QtW.QWidget) -> None:
        if control := self._widget_map.pop(widget, None):
            self.removeWidget(control)

    def update_control_widget(self, current: QtW.QWidget | None) -> None:
        if control := self._widget_map.get(current):
            self.setCurrentWidget(control)
        else:
            self.setCurrentWidget(self._empty_widget)



"""A GUI-independent utility functions to read/write files."""

from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path
from himena.types import WidgetDataModel
from himena._providers import ReaderStore, WriterStore

if TYPE_CHECKING:
    from himena.plugins import ReaderPlugin


def read(
    path: str | Path,
    *,
    plugin: str | None = None,
) -> WidgetDataModel:
    """Read a file as a data model."""
    ins = ReaderStore.instance()
    return ins.run(Path(path), plugin=plugin)


def get_readers(path: str | Path, min_priority: int = 0) -> list[ReaderPlugin]:
    """Get the list of readers that can read the file."""
    ins = ReaderStore.instance()
    return ins.get(Path(path), empty_ok=True, min_priority=min_priority)


def write(
    model: WidgetDataModel,
    path: str | Path,
    *,
    plugin: str | None = None,
) -> None:
    """Write a data model to a file."""
    ins = WriterStore.instance()
    path = Path(path)
    if path.suffix == ".pickle":
        ins.run(model, path, plugin=plugin)
    else:
        ins.run(model, path, plugin=plugin, min_priority=0)
    return None



from typing import Any
from himena.standards.plotting import layout, layout3d


def figure(background_color: Any = "white") -> layout.SingleAxes:
    """Make a single axes layout model.

    Examples
    --------
    ``` python
    from himena.standards import plotting as hplt
    fig = hplt.figure()
    fig.plot([0, 1, 2], [4, 2, 3], color="red")
    fig.show()  # show as a sub-window in the current widget
    ```
    """
    lo = layout.SingleAxes(background_color=background_color)
    return lo


def figure_3d(background_color: Any = "white") -> layout3d.SingleAxes3D:
    """Make a single 3D axes layout model.

    Examples
    --------
    ``` python
    from himena.standards import plotting as hplt
    fig = hplt.figure_3d()
    fig.plot([0, 1, 2], [4, 2, 3], [6, 8, 7], color="red")
    fig.show()  # show as a sub-window in the current widget
    ```
    """
    lo = layout3d.SingleAxes3D(background_color=background_color)
    return lo


def figure_stack(
    *shape,
    background_color: Any = "white",
    multi_dims=None,
) -> layout.SingleStackedAxes:
    if len(shape) == 1 and isinstance(multi_dims, str):
        multi_dims = [multi_dims]
    lo = layout.SingleStackedAxes.fill(*shape, multi_dims=multi_dims)
    lo.background_color = background_color
    return lo


def row(num: int = 1, *, background_color: Any = "white") -> layout.Row:
    """Make a row layout model.

    Examples
    --------
    ``` python
    from himena.standards import plotting as hplt
    row = hplt.row(2)
    row[0].plot([0, 1, 2], [4, 2, 3], color="red")
    row.show()  # show as a sub-window in the current widget
    ```
    """
    lo = layout.Row.fill(num)
    lo.background_color = background_color
    return lo


def column(num: int = 1, *, background_color: Any = "white") -> layout.Column:
    """Make a column layout model.

    Parameters
    ----------
    num : int, optional
        Number of columns, by default 1

    Examples
    --------
    ``` python
    from himena.standards import plotting as hplt
    col = hplt.column(3)
    col[0].plot([0, 1, 2], [4, 2, 3], color="blue")
    col.show()  # show as a sub-window in the current widget
    ```
    """
    lo = layout.Column.fill(num)
    lo.background_color = "white"
    return lo


def grid(
    rows: int = 1, cols: int = 1, *, background_color: Any = "white"
) -> layout.Grid:
    """Make a grid layout model.

    Parameters
    ----------
    rows : int, optional
        Number of rows, by default 1
    cols : int, optional
        Number of columns, by default 1

    Examples
    --------
    ``` python
    from himena.standards import plotting as hplt
    grd = hplt.grid(2, 3)
    grd[0, 0].plot([0, 1, 2], [4, 2, 3], color="green")
    grd.show()  # show as a sub-window in the current widget
    ```
    """
    lo = layout.Grid.fill(rows, cols)
    lo.background_color = "white"
    return lo



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy import QtWidgets as QtW, QtGui, QT6, QtCore
from qtpy.QtCore import Qt

from himena.qt.settings._theme import QThemePanel
from himena.qt.settings._plugins import QPluginListEditor
from himena.qt.settings._startup_commands import QStartupCommandsPanel
from himena.qt.settings._configs import QPluginConfigs
from himena.qt.settings._keybind_edit import QKeybindEdit

if TYPE_CHECKING:
    from himena.widgets import MainWindow


class QSettingsDialog(QtW.QDialog):
    """The dialog used for settings."""

    def __init__(self, ui: MainWindow) -> None:
        super().__init__(ui._backend_main_window)
        self._ui = ui
        self.setWindowTitle("Settings")
        self.resize(600, 400)
        layout = QtW.QHBoxLayout(self)

        layout_left = QtW.QVBoxLayout()
        layout_left.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(layout_left)

        self._list = QtW.QListWidget(self)
        self._list.setFixedWidth(150)
        self._list.setFont(QtGui.QFont("Arial", 13))
        self._open_json_btn = QtW.QPushButton("Open JSON")
        self._open_json_btn.clicked.connect(self._open_json)
        self._stack = QtW.QStackedWidget(self)

        self._list.currentRowChanged.connect(
            lambda: self._stack.setCurrentIndex(self._list.currentRow())
        )

        layout_left.addWidget(self._list)
        layout_left.addWidget(self._open_json_btn)
        layout.addWidget(self._stack)

        self._setup_panels()
        self._list.setCurrentRow(0)

    def addPanel(self, name: str, title: str, panel: QtW.QWidget) -> None:
        """Add a panel to the settings dialog, with the corresponding list item."""
        self._list.addItem(name)
        widget = QtW.QWidget(self._stack)
        layout = QtW.QVBoxLayout(widget)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._stack.addWidget(widget)
        layout.addWidget(QTitleLabel(title, 18))
        layout.addWidget(panel)
        return None

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if (
            a0.key() in (Qt.Key.Key_W, Qt.Key.Key_Q)
            and a0.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            self.close()
        return super().keyPressEvent(a0)

    def _setup_panels(self):
        self.addPanel("Apperance", "Color Theme", QThemePanel(self._ui))
        self.addPanel("Plugins", "Plugins", QPluginListEditor(self._ui))
        self.addPanel("Startup", "Startup Commands", QStartupCommandsPanel(self._ui))
        self.addPanel("Keybindings", "Keybindings", QKeybindEdit(self._ui))
        self.addPanel(
            "Configurations", "Plugin Configurations", QPluginConfigs(self._ui)
        )

    def _open_json(self):
        self._ui.read_file(self._ui.app_profile.profile_path())
        self.close()


class QTitleLabel(QtW.QLabel):
    """Label used for titles in the preference dialog."""

    def __init__(self, text: str, size: int) -> None:
        super().__init__()
        self.setText(text)
        self.setFont(QtGui.QFont("Arial", size))

    def paintEvent(self, a0: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        color = foreground_color_role(self.palette())
        painter.setPen(QtGui.QPen(color, 1))
        bottom_left = self.rect().bottomLeft()
        bottom_right = QtCore.QPoint(bottom_left.x() + 300, bottom_left.y())
        painter.drawLine(bottom_left, bottom_right)
        return super().paintEvent(a0)


def foreground_color_role(qpalette: QtGui.QPalette) -> QtGui.QColor:
    if QT6:
        return qpalette.color(
            QtGui.QPalette.ColorGroup.Normal, QtGui.QPalette.ColorRole.Text
        )
    else:
        return qpalette.color(QtGui.QPalette.ColorRole.Foreground)



from __future__ import annotations

import numpy as np
from magicgui.widgets.bases import ValueWidget
from magicgui.types import Undefined
from magicgui.application import use_app
from magicgui.backends._qtpy.widgets import QBaseValueWidget
from himena.qt._qdtypeedit import QNumericDTypeEdit


class QBaseNumericDTypeEdit(QBaseValueWidget):
    def __init__(self, **kwargs) -> None:
        super().__init__(
            QNumericDTypeEdit, "dtype", "set_dtype", "valueChanged", **kwargs
        )


class NumericDTypeEdit(ValueWidget[str]):
    def __init__(self, value=Undefined, **kwargs):
        app = use_app()
        assert app.native
        if value is Undefined:
            value = "float64"
        elif isinstance(value, np.dtype):
            value = str(value)
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=QBaseNumericDTypeEdit,
            **kwargs,
        )



from qtpy import QtCore
from pytestqt.qtbot import QtBot
from himena import MainWindow
from himena.plugins import update_config_context
from himena_builtins.qt.favorites._widget import QFavoriteCommands, QCommandPushButton
from himena_builtins.qt.favorites import FavoriteCommandsConfig

def test_favorite_commands_widget(qtbot: QtBot, himena_ui: MainWindow):
    widget = QFavoriteCommands(himena_ui)
    qtbot.addWidget(widget)
    himena_ui.exec_action("builtins:favorite-commands")
    mime = QtCore.QMimeData()
    mime.setData("text/command-id", b"new-tab")
    assert widget._command_list.count() == 0

    # dropping a new command
    widget._command_list._drop_mime_data(mime)
    assert widget._command_list.count() == 1

    # dropping the existing command should not add a new one
    widget._command_list._drop_mime_data(mime)
    assert widget._command_list.count() == 1

    item = widget._command_list.item(0)
    btn = widget._command_list.itemWidget(item)
    assert isinstance(btn, QCommandPushButton)
    btn._make_context_menu()
    cfg = himena_ui.app_profile.plugin_configs["builtins:favorite-commands"]
    assert cfg["commands"]["value"] == ["new-tab"]

    with update_config_context(FavoriteCommandsConfig, update_widget=True) as cfg:
        cfg.commands = ["close-tab"]



from __future__ import annotations

import inspect
from pathlib import Path
from typing import (
    Callable,
    Generic,
    Hashable,
    Literal,
    TypeVar,
    TYPE_CHECKING,
    overload,
)

from himena.types import (
    WindowState,
    ClipboardDataModel,
    DockArea,
    DockAreaString,
    WindowRect,
    BackendInstructions,
    WidgetClassTuple,
)

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray
    from concurrent.futures import Future
    from himena.app import EventLoopHandler
    from himena.style import Theme
    from himena.widgets._main_window import MainWindow
    from himena.widgets._wrapper import SubWindow, ParametricWindow

_W = TypeVar("_W")  # backend widget type
_T = TypeVar("_T")


class BackendMainWindow(Generic[_W]):  # pragma: no cover
    _himena_main_window: MainWindow[_W]
    _event_loop_handler: EventLoopHandler

    def __init_subclass__(cls) -> None:
        for name in dir(BackendMainWindow):
            if not hasattr(cls, name):
                raise NotImplementedError(f"Method {name} is not implemented.")

    def _update_widget_theme(self, theme: Theme):
        """Update the theme of the main window."""

    def _main_window_rect(self) -> WindowRect:
        """Get the rect of the main window."""

    def _set_main_window_rect(self, rect: WindowRect) -> None:
        """Set the rect of the main window."""

    def _current_tab_index(self) -> int | None:
        """Get the current tab index.

        If there is no tab, return None.
        """

    def _set_current_tab_index(self, i_tab: int) -> None:
        """Update the current tab index."""

    def _tab_hash(self, i_tab: int) -> Hashable:
        """Get a hashable value of the tab at the index."""

    def _tab_hash_for_window(self, widget: _W) -> Hashable:
        """Get a hashable value of the tab containing the window."""

    def _num_tabs(self) -> int:
        """Get the number of tabs."""

    def _current_sub_window_index(self, i_tab: int) -> int | None:
        """Get the current sub window index in the given tab.

        If there is no sub window, or the tab area itself is selected, return None.
        """

    def _set_current_sub_window_index(self, i_tab: int, i_window: int | None) -> None:
        """Update the current sub window index in the given tab.

        if `i_window` is None, the tab area itself will be selected (all the windows
        will be deselected). `i_window` is asserted to be non-negative.
        """

    def _set_control_widget(self, widget: _W, control: _W | None) -> None:
        """Set the control widget for the given sub window widget.

        A control widget appears on the top-right corner of the toolbar, which will be
        used to display the state of the widget, edit the widget efficiently, etc. For
        example, a font size spinbox for a text editor widget.
        """

    def _update_control_widget(self, current: _W | None) -> None:
        """Switch the control widget to another one in the existing ones.

        If None is given, the control widget will be just hidden.
        """

    def _remove_control_widget(self, widget: _W) -> None:
        """Remove the control widget for the given sub window widget from the stack."""

    def _window_state(self, widget: _W) -> WindowState:
        """The state (min, normal, etc.) of the window."""

    def _set_window_state(
        self,
        widget: _W,
        state: WindowState,
        inst: BackendInstructions,
    ) -> None:
        """Update the state of the window.

        The BackendInstructions indicates the animation or other effects to be applied.
        """

    def _tab_title(self, i_tab: int) -> str:
        """Get the title of the tab at the index."""

    def _set_tab_title(self, i_tab: int, title: str) -> None:
        """Update the title of the tab at the index."""

    def _window_title(self, widget: _W) -> str:
        """Get the title of the window."""

    def _set_window_title(self, widget: _W, title: str) -> None:
        """Update the title of the window."""

    def _window_rect(self, widget: _W) -> WindowRect:
        """Get the rectangle relative to the tab area of the window."""

    def _set_window_rect(
        self,
        widget: _W,
        rect: WindowRect,
        inst: BackendInstructions,
    ) -> None:
        """Update the rectangle of the window.

        The BackendInstructions indicates the animation or other effects to be applied.
        """

    def _area_size(self) -> tuple[int, int]:
        """Get the size of the tab area."""

    @overload
    def _open_file_dialog(
        self,
        mode: Literal["r", "d", "w"] = "r",
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: Path | None = None,
    ) -> Path | None: ...
    @overload
    def _open_file_dialog(
        self,
        mode: Literal["rm"],
        extension_default: str | None = None,
        allowed_extensions: list[str] | None = None,
        caption: str | None = None,
        start_path: Path | None = None,
    ) -> list[Path] | None: ...

    def _open_file_dialog(
        self,
        mode,
        extension_default=None,
        allowed_extensions=None,
        caption=None,
        start_path=None,
    ):
        """Open a file dialog."""

    def _request_choice_dialog(
        self,
        title: str,
        message: str,
        choices: list[tuple[str, _T]],
        how: Literal["buttons", "radiobuttons"] = "buttons",
    ) -> _T | None:
        """Request a choice dialog and return the clicked text."""

    def _show_command_palette(self, kind: str) -> None:
        """Show the command palette widget of the given kind."""

    def _exit_main_window(self, confirm: bool = False) -> None:
        """Close the main window (confirm if needed)."""

    def _get_widget_list(self, i_tab: int) -> list[tuple[str, _W]]:
        """Get the list of widgets in the tab."""

    def _del_widget_at(self, i_tab: int, i_window: int) -> None:
        """Delete the `i_window`-th window in the `i_tab`-th tab."""

    def _get_tab_name_list(self) -> list[str]:
        """Get the list of tab names."""

    def _del_tab_at(self, i_tab: int) -> None:
        """Delete the `i_tab`-th tab.

        Backend does not need to close the subwindows one by one (will be done on the
        wrapper side).
        """

    def _rename_window_at(self, i_tab: int, i_window: int) -> None:
        """Start renaming the `i_window`-th window in the `i_tab`-th tab."""

    def add_widget(self, widget: _W, i_tab: int, title: str) -> _W:
        """Add a sub window containing the widget to the tab at the index.

        Return the backend widget.
        """

    def set_widget_as_preview(self, widget: _W):
        """Set the widget state as the preview mode."""

    def add_tab(self, title: str) -> None:
        """Add a empty tab with the title."""

    def add_dock_widget(
        self,
        widget: _W,
        title: str | None,
        area: DockAreaString | DockArea | None = DockArea.RIGHT,
        allowed_areas: list[DockAreaString | DockArea] | None = None,
    ) -> _W:
        """Add a dock widget containing the widget to the main window.

        Return the backend dock widget.
        """

    ### dock widgets ###
    def _dock_widget_visible(self, widget: _W) -> bool:
        """Whether the dock widget is visible."""

    def _set_dock_widget_visible(self, widget: _W, visible: bool) -> None:
        """Update the visibility of the dock widget."""

    def _dock_widget_title(self, widget: _W) -> str:
        """Get the title of the dock widget."""

    def _set_dock_widget_title(self, widget: _W, title: str) -> None:
        """Update the title of the dock widget."""

    def _del_dock_widget(self, widget: _W) -> None:
        """Delete the dock widget."""

    ### others ###
    def show(self, run: bool = False) -> None:
        """Show the main window and run the app immediately if `run` is True"""

    def _list_widget_class(
        self,
        type: str,
    ) -> tuple[list[WidgetClassTuple], type[_W]]:
        """List the available widget classes of the given type.

        The method will return (list of (widget model type, widget_class, priority),
        fallback class)
        """

    def _connect_main_window_signals(self, main_window: MainWindow[_W]):
        """Connect the signal of the backend main window to the callbacks."""

    def _connect_window_events(
        self,
        wrapper: SubWindow[_W],
        backend: _W,
    ):
        """Connect the events between the wrapper sub window and the backend widget."""

    def _update_context(self) -> None:
        """Update the application context."""

    def _clipboard_data(self) -> ClipboardDataModel | None:
        """Get the clipboard data."""

    def _set_clipboard_data(self, data: ClipboardDataModel) -> None:
        """Set the clipboard data."""

    def _screenshot(self, target: str) -> NDArray[np.uint8]:
        """Take a screenshot of the target area."""

    def _process_parametric_widget(self, widget: _W) -> _W:
        """Process a parametric widget so that it can be added to the main window.

        The incoming widget must implements the `get_params` method, which gives the
        dictionary of parameters.
        """

    def _connect_parametric_widget_events(
        self,
        wrapper: ParametricWindow[_W],
        widget: _W,
    ) -> None:
        """Connect the events between the wrapper parametric window and the backend."""

    def _signature_to_widget(
        self,
        sig: inspect.Signature,
        show_parameter_labels: bool = True,
        preview: bool = False,
    ) -> _W:
        """Convert a function signature to a widget that can run it."""

    def _add_widget_to_parametric_window(
        self,
        wrapper: ParametricWindow[_W],
        widget: _W,
        result_as: Literal["below", "right"],
    ) -> None:
        """Add a widget to the parametric window."""

    def _remove_widget_from_parametric_window(
        self,
        wrapper: ParametricWindow[_W],
    ) -> None:
        """Remove a widget from the parametric window."""

    def _move_focus_to(self, widget: _W) -> None:
        """Move the focus to the widget."""

    def _set_status_tip(self, tip: str, duration: float) -> None:
        """Set the status tip of the main window for a duration (sec)."""

    def _show_notification(self, text: str, duration: float) -> None:
        """Show notification for a duration (sec)."""

    def _rebuild_for_runtime(self, new_menus: list[str]) -> None:
        """Register the actions at runtime."""

    def _process_future_done_callback(
        self,
        cb: Callable[[Future], None],
        cb_errored: Callable[[Exception], None],
        **kwargs,
    ) -> Callable[[Future], None]:
        """Wrap the callback of the future done event so that it can be run in the main
        thread."""

    def _set_parametric_widget_busy(self, wrapper: ParametricWindow[_W], busy: bool):
        """Set the parametric widget busy status (disable call button etc)."""

    def _add_job_progress(self, future: Future, desc: str, total: int = 0) -> None:
        """Add a job to the job stack."""

    def _add_whats_this(self, text: str, style: Literal["plain", "markdown", "html"]):
        """Add a what's this text to the main window."""

    def _append_result(self, item: dict[str, object]) -> None:
        """Append a new result to the result stack."""



from __future__ import annotations
from typing import Any

from qtpy import QtCore, QtWidgets as QtW, QtGui
from himena.types import WidgetDataModel
from himena.consts import StandardType
from himena.standards.model_meta import ListMeta
from himena.workflow import ProgrammaticMethod
from himena.qt._utils import get_main_window
from dataclasses import asdict, is_dataclass


class QResultStack(QtW.QTableWidget):
    """A special widget for displaying measurement results."""

    __himena_widget_id__ = "QResultStack"

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSelectionBehavior(QtW.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setVerticalScrollMode(QtW.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.setHorizontalScrollMode(QtW.QAbstractItemView.ScrollMode.ScrollPerPixel)
        self.horizontalHeader().hide()
        self._items: list[dict[str, Any]] = []  # the actual python objects

    def append_result(self, item: dict[str, Any]) -> None:
        """Append a new result to the stack."""
        # coerce item to a dict-like object
        if isinstance(item, dict):
            pass
        elif is_dataclass(item):
            item = asdict(item)
        else:
            item = dict(item)
        self.insertRow(self.rowCount())
        self.setRowHeight(self.rowCount(), 18)
        add_new_column = len(item) > self.columnCount()
        if add_new_column:
            self.setColumnCount(len(item))
        flags = QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        for column, (key, value) in enumerate(item.items()):
            label = QtW.QLabel(f"<b><font color='#808080'>{key}:</font></b> {value!r}")
            label.setContentsMargins(5, 2, 5, 2)
            table_item = QtW.QTableWidgetItem()
            table_item.setToolTip(f"{key}: {value!r}")
            table_item.setFlags(flags)
            model_index = self.model().index(self.rowCount() - 1, column)
            self.setIndexWidget(model_index, label)
            self.setItem(self.rowCount() - 1, column, table_item)
        self._items.append(item)

    def update_model(self, model: WidgetDataModel) -> None:
        """Update the model with new data."""
        vals = model.value
        if not hasattr(vals, "__iter__"):
            raise TypeError(f"Expected an iterable, got {type(vals)}")
        self.clear()
        for val in vals:
            self.append_result(val)

    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self._items,
            type=StandardType.RESULTS,
            title="Results",
            metadata=ListMeta(selections=self.selections()),
            workflow=ProgrammaticMethod().construct_workflow(),
        )

    def model_type(self) -> StandardType:
        return StandardType.RESULTS

    def size_hint(self) -> tuple[int, int]:
        """Return the size hint for the widget."""
        return 320, 320

    def selections(self) -> list[int]:
        """Return the selected row indices."""
        return sorted({i.row() for i in self.selectedIndexes()})

    def _make_context_menu(self) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        copy_action = menu.addAction(
            "Copy", lambda: self._copy_items(self.selections())
        )
        copy_action.setToolTip("Copy selected items to clipboard")
        select_action = menu.addAction(
            "Select Rows With Same Keys", lambda: self._select_rows_with_same_keys()
        )
        select_action.setToolTip("Select all rows with the same keys")
        return menu

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        """Handle mouse press events."""
        if event.buttons() & QtCore.Qt.MouseButton.RightButton:
            menu = self._make_context_menu()
            self.selectRow(self.rowAt(event.y()))
            menu.exec(event.globalPos())
        else:
            super().mousePressEvent(event)

    def _copy_items(self, indices: list[int]) -> None:
        """Copy selected items to clipboard."""
        if not indices:
            return
        items = [self._items[i] for i in indices]
        ui = get_main_window(self)
        model = WidgetDataModel(value=items, type=StandardType.RESULTS)
        ui.set_clipboard(text=repr(items), internal_data=model)

    def _select_rows_with_same_keys(self) -> None:
        """Select rows with the same keys."""
        sels = self.selections()
        if not sels:
            return
        item_ref = self._items[sels[0]]
        self.clearSelection()
        selected = []
        for i, item in enumerate(self._items):
            if item.keys() == item_ref.keys():
                selected.append(i)
        for i in selected:
            for j in range(self.columnCount()):
                self.item(i, j).setSelected(True)

    def keyPressEvent(self, e):
        if (
            e.key() == QtCore.Qt.Key.Key_C
            and e.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
        ):
            self._copy_items(self.selections())
            return
        return super().keyPressEvent(e)



from __future__ import annotations

from dataclasses import dataclass, field
import timeit
from typing import Generic, TypeVar
import warnings
from himena.types import WidgetDataModel
from himena.plugins import AppActionRegistry
from himena.utils.misc import is_subtype
from himena.workflow import (
    CommandExecution,
    parse_parameter,
    ProgrammaticMethod,
    UserModification,
)

_T = TypeVar("_T")


@dataclass
class Modifications(Generic[_T]):
    """Class that tracks modifications to a SubWindow."""

    initial_value: _T | None = field(default=None)
    initial_time: float = field(default_factory=timeit.default_timer)
    track_enabled: bool = field(default=False)

    def update_workflow(self, model: WidgetDataModel) -> None:
        """Update the workflow with a modification step."""
        if self.initial_value is None or not self.track_enabled:
            return None  # No need to update the workflow
        if isinstance(model.workflow.last(), (UserModification, ProgrammaticMethod)):
            # Already has a untrackable modification, so we don't need to do anything
            return
        _reg = AppActionRegistry.instance()
        diff = None
        mod_tracker = None
        for _type in _reg._modification_trackers:
            if is_subtype(model.type, _type):
                # Look for the modification tracker
                mod_tracker = _reg._modification_trackers[_type]
                break
        if mod_tracker is None:
            warnings.warn(
                f"Modification tracking not available for {model.type}.",
                UserWarning,
                stacklevel=2,
            )
        else:
            try:
                diff = mod_tracker(self.initial_value, model.value)
            except Exception as e:
                # If the diff fails, warn the user and skip the tracking to avoid crashing
                warnings.warn(
                    f"Modification tracking failed for {model.type}: {e}",
                    UserWarning,
                    stacklevel=2,
                )
                diff = None
        if diff:
            model_param = parse_parameter("model", model)
            model.workflow = model.workflow.with_step(
                CommandExecution(
                    command_id=diff.command_id,
                    contexts=[model_param[0]],
                    parameters=[
                        parse_parameter(name, value)[0]
                        for name, value in diff.with_params.items()
                    ],
                    execution_time=timeit.default_timer() - self.initial_time,
                )
            )
        else:
            model.workflow = model.workflow.with_step(
                UserModification(original=model.workflow.last_id())
            )



from __future__ import annotations

from pathlib import Path
from typing import TypeVar, Generic

_T = TypeVar("_T")


class HistoryContainer(Generic[_T]):
    def __init__(self, max_size: int = 20):
        self._hist: list[_T] = []
        self._max_size = max_size

    def add(self, item: _T) -> None:
        """Add item to the history."""
        self._hist.append(item)
        if len(self._hist) > self._max_size:
            self._hist.pop(0)

    def get(self, num: int) -> _T | None:
        """Get the item at the given index if exists."""
        if len(self._hist) > num:
            return self._hist[num]
        return None

    def get_from_last(self, num: int) -> _T | None:
        """Get the item at the given index from the last if exists."""
        if len(self._hist) >= num:
            return self._hist[-num]
        return None

    def pop_last(self) -> _T | None:
        """Pop the last item if exists."""
        if self._hist:
            return self._hist.pop()
        return None

    def len(self) -> int:
        return len(self._hist)

    def __len__(self) -> int:
        return self.len()


class FileDialogHistoryDict:
    """List of starting directories for file dialogs."""

    def __init__(self):
        self._dict: dict[str, Path] = {}

    def update(self, group_name: str, value: Path) -> None:
        self._dict[group_name] = value

    def get_path(self, group_name: str, filename: str | None = None) -> Path:
        d = self._dict.get(group_name, Path.cwd())
        if filename:
            return d / filename
        while not d.exists():
            d = d.parent
            if d == Path.home() or d == Path("."):
                break
        return d



import re
import html

HTML_PATTERN = re.compile(r"<.*?>")
HTML_BLOCK_PATTERN = re.compile(r"<html>.*?</html>", re.DOTALL)
HEADER_PATTERN = re.compile(r"<head>.*?</head>", re.DOTALL)
NEWLINE_PATTERN = re.compile(r"<br\s*/?>", re.IGNORECASE)
P_PATTERN = re.compile(r"<p.*?>(.*?)</p>", re.DOTALL)


def html_to_plain_text(html_text: str) -> str:
    """Convert HTML text to plain text."""
    html_text = html_text.replace("\n", "")
    if html_block_match := HTML_BLOCK_PATTERN.search(html_text):
        html_text = html_block_match.group(0)
    html_text = NEWLINE_PATTERN.sub("\n", html_text)
    html_text = "\n".join(m.group(1) for m in P_PATTERN.finditer(html_text))
    return html.unescape(HTML_PATTERN.sub("", HEADER_PATTERN.sub("", html_text)))



from __future__ import annotations
from typing import Iterable, TYPE_CHECKING
from app_model import Application
from app_model.types import MenuItem

if TYPE_CHECKING:
    from app_model.types import MenuOrSubmenu, Action


def collect_commands(
    app: Application,
    menu_items: list[MenuOrSubmenu],
    exclude: Iterable[str],
) -> list[Action]:
    commands = []
    for item in menu_items:
        if isinstance(item, MenuItem):
            if item.command.id not in exclude:
                commands.append(item.command)
        else:
            commands.extend(
                collect_commands(app, app.menus.get_menu(item.submenu), exclude)
            )
    return commands
